import pyarrow.dataset as ds
import csv
import numpy as np
from tgb.utils.stats import analyze_csv
import pandas as pd
from tqdm import tqdm


def collect_csv(dir_name="software"):
    dataset = ds.dataset(dir_name, format="csv")
    df = dataset.to_table().to_pandas()
    df.to_csv(dir_name + ".csv", index=True)


def reorder_column(fname: str, outname: str):
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "source", "target", "weight"]
        write.writerow(fields)
        line_count = 0
        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            for row in csv_reader:
                if line_count == 0:  # header
                    line_count += 1
                else:
                    # edgeid, SourceId,TargetId,Weight,Timestamp
                    src = row[1]
                    dst = row[2]
                    w = row[3]
                    ts = row[4]
                    write.writerow([ts, src, dst, w])
                    line_count += 1


def sort_edgelist(fname: str, outname: str):
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "source", "target", "weight"]
        write.writerow(fields)
        line_count = 0
        ts_list = []
        line_list = []

        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            for row in csv_reader:
                if line_count == 0:  # header
                    line_count += 1
                else:
                    ts = int(row[0])
                    src = row[1]
                    dst = row[2]
                    w = row[3]
                    ts_list.append(ts)
                    line_list.append([ts, src, dst, w])
                    # write.writerow([ts, src, dst, w])
                    line_count += 1

        ts_list = np.array(ts_list)
        idx = np.argsort(ts_list)
        idx = idx.tolist()

        line_list_out = []
        for i in idx:
            line_list_out.append(line_list[i])
        for line in line_list_out:
            write.writerow(line)


def count_degree(fname: str):
    node_counts = {}
    line_count = 0
    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        for row in csv_reader:
            if line_count == 0:  # header
                line_count += 1
            else:
                ts = int(row[0])
                src = row[1]
                dst = row[2]
                w = row[3]
                if src not in node_counts:
                    node_counts[src] = 1
                else:
                    node_counts[src] += 1

                if dst not in node_counts:
                    node_counts[dst] = 1
                else:
                    node_counts[dst] += 1
                line_count += 1
    return node_counts


def reduce_edgelist(fname: str, outname: str, node10_id: dict):
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "source", "target", "weight"]
        write.writerow(fields)
        line_count = 0

        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            for row in csv_reader:
                if line_count == 0:  # header
                    line_count += 1
                else:
                    ts = int(row[0])
                    src = row[1]
                    dst = row[2]
                    w = row[3]
                    if (src in node10_id) and (dst in node10_id):
                        write.writerow([ts, src, dst, w])
                    line_count += 1

"""
function for review
"""
def csv_process_review(
    fname: str,
    outname: str = "review.csv",
) -> pd.DataFrame:
    r"""
    used for processing review dataset, helper function, not used in actual dataloading
    input .csv file format should be: timestamp, node u, node v, attributes
    Parameters:
        fname: the path to the raw data
    Returns:
        df: a pandas dataframe containing the edgelist data
        feat_l: a numpy array containing the node features
        node_ids: a dictionary mapping node id to integer
    """
    src_ids = {}
    dst_ids = {}
    src_ctr = 0
    dst_ctr = 0

    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        idx = 0
        """
        ts,source,target,weight
        929232000,137139,30122,5.0
        930787200,129185,175070,2.0
        931824000,246213,30122,2.0
        """
        for row in tqdm(csv_reader):
            if idx == 0:
                idx += 1
                continue
            else:
                ts = int(row[0])
                src = row[1]
                dst = row[2]
                if src not in src_ids:
                    src_ids[src] = src_ctr
                    src_ctr += 1
                if dst not in dst_ids:
                    dst_ids[dst] = dst_ctr
                    dst_ctr += 1
                w = float(row[3])
    
    #! ensure that source and destination nodes are unique and non-overlapping
    src_ctr += 1
    dst_ids = {k:v+src_ctr for k,v in dst_ids.items()}

    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts","source","target","weight"]
        write.writerow(fields)

        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            idx = 0
            """
            ts,source,target,weight
            929232000,137139,30122,5.0
            930787200,129185,175070,2.0
            931824000,246213,30122,2.0
            """
            for row in tqdm(csv_reader):
                if idx == 0:
                    idx += 1
                    continue
                else:
                    ts = int(row[0])
                    src = src_ids[row[1]]
                    dst = dst_ids[row[2]]
                    w = float(row[3])
                write.writerow([ts,src,dst,w])


def main():
    # # collect csv
    # # collect_csv(dir_name = "software")
    # collect_csv(dir_name="books")
    # # collect_csv(dir_name = "electronics")

    # # #* reorder column
    # # fname = "electronics.csv"
    # # outname = "amazonreview_edgelist.csv"
    # # reorder_column(fname,
    # #                outname)

    # # #* sort edgelist
    # # fname = "amazonreview_edgelist.csv"
    # # outname = "amazonreview_edgelist_sort.csv"
    # # sort_edgelist(fname,
    # #               outname)

    # fname = "amazonreview_edgelist_reduce.csv"
    # analyze_csv(fname)

    # # fname = "amazonreview_edgelist.csv"
    # # node_counts = count_degree(fname)
    # # node10_id = {}
    # # for node in node_counts:
    # #     if node_counts[node] > 10:
    # #         node10_id[node] = node_counts[node]

    # # outname = "amazonreview_edgelist_reduce.csv"
    # # reduce_edgelist(fname,
    # #                 outname,
    # #                 node10_id)

    csv_process_review("tgbl-review_edgelist_v2.csv", "review.csv")

    


if __name__ == "__main__":
    main()



import csv
import datetime
import glob, os


def load_time_csv_raw(fname):
    r"""
    load from the raw data and retrieve, timestamp, head, tail, relation, time_rel
    convert all dates into unix timestamps
    """
    out_dict = {}
    first_row = True
    num_lines = 0
    #? timestamp,head,tail,relation_type,time_rel_type
    #* +1999-01-01T00:00:00Z,Q31,Q4916,P38,P580
    error_ctr = 0
    with open(fname, 'r') as f:
        reader = csv.reader(f, delimiter =',')

        for row in reader: 
            if first_row:
                first_row = False
                continue
            date = row[0][0:11]
            head = row[1]
            tail = row[2]
            relation_type = row[3]
            time_rel = row[4]

            if (len(date) == 0):
                continue
            
            if ("None" in date or "None" in head or "None" in tail or "None" in relation_type):
                continue
            else:
                TIME_FORMAT = "%Y"
                #* only keep track of year in positive BC
                if (date[0] == "+"):
                    ts = int(date[1:5])
                else:
                    continue

                #* no scifi for knowledge graphs 
                if (ts > 2024):
                    continue

                num_lines += 1
                if (ts in out_dict):
                    if (head, tail, relation_type, time_rel) in out_dict[ts]:
                        out_dict[ts][(head, tail, relation_type, time_rel)] += 1
                    else:
                        out_dict[ts][(head, tail, relation_type, time_rel)] = 1
                else:
                    out_dict[ts] = {}
                    out_dict[ts][(head, tail, relation_type, time_rel)] = 1
    return out_dict, num_lines


def write2csv(outname: str, 
              out_dict: dict,):
    r"""
    Write the dictionary to a csv file
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['ts', 'head', 'tail', 'relation_type','time_rel_type'])
        dates = list(out_dict.keys())
        dates.sort()
        for date in dates:
            for edge in out_dict[date]:
                head = edge[0]
                tail = edge[1]
                relation_type = edge[2]
                time_rel = edge[3]
                row = [date, head, tail, relation_type, time_rel]
                writer.writerow(row)


def update_dict(total_dict, new_dict):
    r"""
    Update the total_dict with new_dict
    """
    for key in new_dict:
        if key in total_dict:
            for edge in new_dict[key]:
                if edge in total_dict[key]:
                    total_dict[key][edge] += new_dict[key][edge]
                else:
                    total_dict[key][edge] = new_dict[key][edge]
        else:
            total_dict[key] = new_dict[key]
    return total_dict


def retrieve_all_entities(total_dict):
    r"""
    retrieve the entities from all edges of the total dictionary

    Parameters:
        total_dict: dictionary of all edges, {ts: {edge: count}}
    """
    node_dict = {}
    for key in total_dict:
        for edge in total_dict[key]:
            head = edge[0]
            tail = edge[1]
            if head not in node_dict:
                node_dict[head] = 1
            else:
                node_dict[head] += 1
            if tail not in node_dict:
                node_dict[tail] = 1
            else:
                node_dict[tail] += 1
    return node_dict


def writenode2csv(outname: str, 
              out_dict: dict,):
    r"""
    Write the dictionary to a csv file
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['entity', 'occurrences'])
        for node in out_dict:
            row = [node, out_dict[node]]
            writer.writerow(row) 


def main():

    #! when timestamps overlap can't update dictionary

    total_lines = 0
    total_edge_dict = {}

    #1. find all files with .txt in the folder
    total_edge_file = "tkgl-wikidata_edgelist.csv"
    for file in glob.glob("*.csv"):
        print (file)
        edge_dict, num_lines = load_time_csv_raw(file)
        print ("processed ", num_lines, " lines")
        total_lines += num_lines
        update_dict(total_edge_dict, edge_dict)
    print ("processed a total of ", total_lines, " lines")
    node_dict = retrieve_all_entities(total_edge_dict)
    writenode2csv("wiki_entities.csv", node_dict)
    write2csv(total_edge_file, total_edge_dict)



if __name__ == "__main__":
    main()


import timeit
from tgb.linkproppred.negative_generator import NegativeEdgeGenerator
from tgb.linkproppred.dataset_pyg import PyGLinkPropPredDataset



def main():
    r"""
    Generate negative edges for the validation or test phase
    """
    print("*** Negative Sample Generation ***")

    # setting the required parameters
    num_neg_e_per_pos = 1000 # more than half the nodes in the graph
    neg_sample_strategy = "hist_rnd" #"rnd"
    rnd_seed = 42


    name = "tgbl-uci"
    dataset = PyGLinkPropPredDataset(name=name, root="datasets")
    train_mask = dataset.train_mask
    val_mask = dataset.val_mask
    test_mask = dataset.test_mask
    data = dataset.get_TemporalData()

    data_splits = {}
    data_splits['train'] = data[train_mask]
    data_splits['val'] = data[val_mask]
    data_splits['test'] = data[test_mask]

    # Ensure to only sample actual destination nodes as negatives.
    min_dst_idx, max_dst_idx = int(data.dst.min()), int(data.dst.max())

    # After successfully loading the dataset...
    if neg_sample_strategy == "hist_rnd":
        historical_data = data_splits["train"]
    else:
        historical_data = None

    neg_sampler = NegativeEdgeGenerator(
        dataset_name=name,
        first_dst_id=min_dst_idx,
        last_dst_id=max_dst_idx,
        num_neg_e=num_neg_e_per_pos,
        strategy=neg_sample_strategy,
        rnd_seed=rnd_seed,
        historical_data=historical_data,
    )

    # generate evaluation set
    partial_path = "./"
    # generate validation negative edge set
    start_time = timeit.default_timer()
    split_mode = "val"
    print(
        f"INFO: Start generating negative samples: {split_mode} --- {neg_sample_strategy}"
    )
    neg_sampler.generate_negative_samples(
        data=data_splits[split_mode], split_mode=split_mode, partial_path=partial_path
    )
    print(
        f"INFO: End of negative samples generation. Elapsed Time (s): {timeit.default_timer() - start_time: .4f}"
    )

    # generate test negative edge set
    start_time = timeit.default_timer()
    split_mode = "test"
    print(
        f"INFO: Start generating negative samples: {split_mode} --- {neg_sample_strategy}"
    )
    neg_sampler.generate_negative_samples(
        data=data_splits[split_mode], split_mode=split_mode, partial_path=partial_path
    )
    print(
        f"INFO: End of negative samples generation. Elapsed Time (s): {timeit.default_timer()- start_time: .4f}"

    )


if __name__ == "__main__":
    main()



import csv
from tqdm import tqdm
from os import listdir
from tgb.utils.stats import analyze_csv


def find_filenames(path_to_dir):
    r"""
    find all files in a folder
    Parameters:
        path_to_dir (str): path to the directory
    """
    filenames = listdir(path_to_dir)
    return filenames


def combine_edgelist_edgefeat2subreddits(edgefname, featfname, outname):
    """
    combine edgelist and edge features
    'ts', 'src', 'subreddit', 'num_words', 'score'
    """
    line_idx = 0

    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "src", "subreddit", "num_words", "score"]
        write.writerow(fields)
        sub_id = 0
        edgelist = open(edgefname, "r")
        edgefeat = open(featfname, "r")
        edgelist.readline()
        edgefeat.readline()

        while True:
            #'ts', 'src', 'dst', 'edge_id'
            edge_line = edgelist.readline()
            edge_line = edge_line.split(",")
            if len(edge_line) < 4:
                break
            edge_id = int(edge_line[3])
            ts = int(edge_line[0])
            src = int(edge_line[1])

            #'edge_id', 'subreddit', 'num_characters', 'num_words', 'score', 'edited_flag'
            feat_line = edgefeat.readline()
            feat_line = feat_line.split(",")
            edge_id_feat = int(feat_line[0])
            subreddit = feat_line[1]
            num_words = int(feat_line[3])
            score = int(feat_line[4])

            if edge_id != edge_id_feat:
                print("edge_id != edge_id_feat")
                print(edge_id)
                print(edge_id_feat)
                break

            write.writerow([ts, src, subreddit, num_words, score])
            line_idx += 1
    print("processed", line_idx, "lines")


def filter_subreddits(fname):
    """
    check the frequency of subreddits
    """
    subreddit_count = {}
    node_count = {}
    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        line_count = 0
        # ts, src, subreddit, num_words, score
        for row in csv_reader:
            if line_count == 0:
                line_count += 1
            else:
                ts = row[0]
                src = row[1]
                subreddit = row[2]
                if subreddit not in subreddit_count:
                    subreddit_count[subreddit] = 1
                else:
                    subreddit_count[subreddit] += 1
                if src not in node_count:
                    node_count[src] = 1
                else:
                    node_count[src] += 1
    return subreddit_count, node_count


def clean_edgelist(fname, node_counts, outname, threshold=1000):
    """
    helper function for filtering out low frequency nodes
    """
    node_dict = {}

    for node in node_counts:
        if node_counts[node] >= threshold:
            node_dict[node] = 1

    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "user", "subreddit", "num_words", "score"]
        write.writerow(fields)

        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            # ts, src, subreddit, num_words, score
            for row in csv_reader:
                if line_count == 0:
                    line_count += 1
                else:
                    ts = row[0]
                    src = row[1]
                    subreddit = row[2]
                    num_words = int(row[3])
                    score = int(row[4])
                    if src in node_dict:
                        write.writerow([ts, src, subreddit, num_words, score])


def clean_edgelist_reddits(fname, reddit_counts, outname, threshold=50):
    """
    helper function for filtering out low frequency subreddits
    """
    reddit_dict = {}

    for reddit in reddit_counts:
        if reddit_counts[reddit] >= threshold:
            reddit_dict[reddit] = 1
    print ("there remains, ", len(reddit_dict), " subreddits")

    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "user", "subreddit", "num_words", "score"]
        write.writerow(fields)
        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            # ts, src, subreddit, num_words, score
            for row in csv_reader:
                if line_count == 0:
                    line_count += 1
                else:
                    ts = row[0]
                    src = row[1]
                    subreddit = row[2]
                    num_words = int(row[3])
                    score = int(row[4])
                    if subreddit in reddit_dict:
                        write.writerow([ts, src, subreddit, num_words, score])


def remove_missing_user(fname, outname):
    """
    remove all lines that are missing the user
    """

    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "user", "subreddit", "num_words", "score"]
        write.writerow(fields)
        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            # ts, src, subreddit, num_words, score
            for row in csv_reader:
                if line_count == 0:
                    line_count += 1
                else:
                    ts = row[0]
                    src = int(row[1])
                    subreddit = row[2]
                    num_words = int(row[3])
                    score = int(row[4])
                    if src != -1:
                        write.writerow([ts, src, subreddit, num_words, score])


def generate_daily_node_labels(
    fname: str,
    outname: str,
):
    r"""
    function for generating daily node labels then can be used for aggregation
    """

    day_dict = {}  # store the weights of genres on this day
    prev_t = -1
    DAY_IN_SEC = 86400
    # WEEK_IN_SEC = 604800
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "user", "subreddit", "weight"]
        write.writerow(fields)

        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            # ts, src, subreddit, num_words, score
            for row in csv_reader:
                if line_count == 0:
                    line_count += 1
                else:
                    ts = int(row[0])
                    user_id = row[1]
                    subreddit = row[2]

                    if line_count == 1:
                        prev_t = ts

                    if (prev_t + DAY_IN_SEC) < ts:
                        #! user,time,genre,weight  # genres = # of weights
                        out = [user_id, ts]
                        for subreddit, w in day_dict.items():
                            write.writerow(out + [subreddit] + [w])
                        prev_t = ts
                        day_dict = {}
                    else:
                        if subreddit not in day_dict:
                            day_dict[subreddit] = 1
                        else:
                            day_dict[subreddit] += 1
                    line_count += 1


#! note that the edgelist are not sorted by users then by time, should keep multiple users when aggregating
def generate_aggregate_labels(fname: str, outname: str, days: int = 7):
    """
    aggregate the genres over a number of days,  as specified by days
    prediction should always be at the first second of the day
    #! daily labels are always shifted by 1 day
    """

    ts_prev = 0

    DAY_IN_SEC = 86400
    timespan = days * DAY_IN_SEC

    user_dict = {}

    # ts, src, subreddit, num_words, score
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "user", "subreddit", "weight"]
        write.writerow(fields)

        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            # ts, src, subreddit, num_words, score
            for row in csv_reader:
                if line_count == 0:
                    line_count += 1
                else:
                    ts = int(row[0])
                    user = row[1]
                    subreddit = row[2]
                    w = int(row[3])
                    if line_count == 1:
                        ts_prev = ts

                    if (ts - ts_prev) > timespan:
                        for user in user_dict:
                            total = sum(user_dict[user].values())
                            subreddit_dict = {
                                k: v / total for k, v in user_dict[user].items()
                            }
                            for subreddit, w in subreddit_dict.items():
                                write.writerow(
                                    [ts_prev + DAY_IN_SEC, user, subreddit, w]
                                )
                        user_dict = {}
                        ts_prev = ts_prev + DAY_IN_SEC  #! move label to the next day
                    else:
                        if user in user_dict:
                            if subreddit in user_dict[user]:
                                user_dict[user][subreddit] += w
                            else:
                                user_dict[user][subreddit] = w
                        else:
                            user_dict[user] = {}
                            user_dict[user][subreddit] = w
                    line_count += 1


def main():
    # #? see redditcomments.py for the extraction from the raw files

    #! combine edgelist and edge feat file check if the edge_id matches
    # edgefname = "redditcomments_edgelist_2008_2010.csv"
    # featfname = "redditcomments_edgefeat_2008_2010.csv"
    # outname = "subreddits_edgelist.csv"
    # combine_edgelist_edgefeat2subreddits(edgefname, featfname, outname)

    #! remove all edges missing user
    # fname = "subreddits_edgelist.csv"
    # outname = "subreddits_edgelist_filtered.csv"
    # remove_missing_user(fname,
    #                     outname)

    #! should clean subreddits first, frequency count of reddits
    # fname = "subreddits_edgelist.csv"
    # outname = "subreddits_edgelist_filter.csv"
    # subreddit_count, node_count = filter_subreddits(fname)
    # threshold = 1000 #200 #100
    # clean_edgelist_reddits(fname, subreddit_count, outname, threshold=threshold)


    #! filter out nodes with low frequency frequency count of nodes
    # fname = "subreddits_edgelist.csv"
    # outname = "subreddits_edgelist_clean.csv"
    # subreddit_count, node_count = filter_subreddits(fname)
    # threshold = 1000
    # clean_edgelist(fname, node_count, outname, threshold=threshold)
    # print ("finish cleaning")

    #! generate aggregate labels, the label for each day is shifted by 1 day as it uses the edges from today
    # fname = "subreddits_edgelist.csv"
    # outname = "subreddits_node_labels.csv"
    # generate_aggregate_labels(fname, outname, days=7)

    #! analyze the extracted csv
    fname = "subreddits_edgelist.csv"
    analyze_csv(fname)


    
    # #! generate daily node labels
    # outname = 'subreddits_daily_labels.csv'
    # fname = "subreddits_edgelist.csv"
    # generate_daily_node_labels(fname,outname)


if __name__ == "__main__":
    main()



import dateutil.parser as dparser
import csv
import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm
from os import listdir
from datetime import datetime

def find_csv_filenames(path_to_dir, suffix=".csv"):
    r"""
    find all csv files in a directory
    Parameters:
        path_to_dir (str): path to the directory
                suffix (str): suffix of the file
    """
    filenames = listdir(path_to_dir)
    return [filename for filename in filenames if filename.endswith(suffix)]


def flight2edgelist(
    fname,
    outname,
    node_dict=None,
):
    """
    process all rows into
    Day, src, dst, callsign, number, icao24, registration, typecode
    and save it as an edgelist file
    """
    miss_node_lines = 0

    skip_lines = 0
    print("processing ", outname)
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = [
            "day",
            "src",
            "dst",
            "callsign",
            "number",
            "icao24",
            "registration",
            "typecode",
        ]
        write.writerow(fields)

        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            # callsign,number,icao24,registration,typecode,origin,destination,firstseen,lastseen,day,latitude_1,longitude_1,altitude_1,latitude_2,longitude_2,altitude_2
            for row in csv_reader:
                if line_count == 0:
                    line_count += 1
                else:
                    out = []
                    callsign = row[0]
                    number = row[1]
                    icao24 = row[2]
                    registration = row[3]
                    typecode = row[4]
                    src = row[5]
                    if src == "":
                        skip_lines += 1
                        continue
                    dst = row[6]
                    if dst == "":
                        skip_lines += 1
                        continue

                    if node_dict is not None:
                        if src not in node_dict:
                            miss_node_lines += 1
                            continue
                        if dst not in node_dict:
                            miss_node_lines += 1
                            continue
                    day = row[9]
                    day = day[0:10]

                    out.append(day)
                    out.append(src)
                    out.append(dst)
                    out.append(callsign)
                    out.append(number)
                    out.append(icao24)
                    out.append(registration)
                    out.append(typecode)
                    write.writerow(out)
                    line_count += 1
        print(f"Processed {line_count} lines.")
        print(f"Skipped {skip_lines} lines.")
        print(f"missing node {miss_node_lines} lines.")
    return line_count, skip_lines, miss_node_lines


def load_icao_airports(fname="airport_codes.csv"):
    airports_continent = {}
    airports_country = {}

    edgelist = open(fname, "r")
    lines = list(edgelist.readlines())
    edgelist.close()
    # date u  v  w
    # find how many timestamps there are

    for i in range(0, len(lines)):
        line = lines[i]
        values = line.split(",")
        icao = values[0]
        continent = values[4]
        country = values[5]
        airports_continent[icao] = continent
        airports_country[icao] = country
    return airports_continent, airports_country


def merge_edgelist(input_names: str, in_dir: str, outname: str):
    """
    merge a list of edgefiles into one file
    """
    line_count = 0
    total = 0
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["day", "src", "dst", "callsign", "typecode"]
        write.writerow(fields)
        for csv_name in tqdm(input_names):
            in_name = in_dir + csv_name
            line_count = 0
            with open(in_name, "r") as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=",")
                for row in csv_reader:
                    if line_count == 0:  # header
                        line_count += 1
                    else:
                        # Day, src, dst, callsign, number, icao24, registration, typecode
                        day = row[0]
                        src = row[1]
                        dst = row[2]
                        callsign = row[3]
                        typecode = row[-1]
                        out = [day, src, dst, callsign, typecode]
                        write.writerow(out)
                        total += 1


def clean_node_feat(in_file, outname):
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = [
            "airport_code",
            "type",
            "continent",
            "iso_region",
            "longitude",
            "latitude",
        ]
        write.writerow(fields)
        idx = 0
        with open(in_file, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            for row in csv_reader:
                if idx == 0:
                    idx += 1
                    continue
                else:
                    # ident,type,name,elevation_ft,continent,iso_country,iso_region,municipality,gps_code,iata_code,local_code,coordinates
                    airport_code = row[0]
                    type = row[1]
                    continent = row[4]
                    iso_region = row[6]
                    longitude = float(row[-1].split(",")[0])
                    latitude = float(row[-1].split(",")[1])
                    out = [
                        airport_code,
                        type,
                        continent,
                        iso_region,
                        longitude,
                        latitude,
                    ]
                    idx += 1
                    write.writerow(out)



def sort_edgelist(in_file, outname):
    """
    sort the edges by day
    """
    TIME_FORMAT = "%Y-%m-%d"
    row_dict = {} #{day: {row: row}}
    line_idx = 0
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["day", "src", "dst", "callsign", "typecode"]
        write.writerow(fields)
        with open(in_file, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            for row in csv_reader:
                if line_idx == 0:  # header
                    line_idx += 1
                    continue
                day = row[0]
                ts = datetime.strptime(day, TIME_FORMAT)
                ts = ts.timestamp()
                if ts not in row_dict:
                    row_dict[ts] = {}
                    row_dict[ts][line_idx] = row
                else:
                    row_dict[ts][line_idx] = row
                line_idx += 1
        
        for ts in sorted(row_dict.keys()):
            for idx in row_dict[ts].keys():
                row = row_dict[ts][idx]
                write.writerow(row)


def date2ts(date_str: str) -> float:
    r"""
    convert date string to timestamp
    """
    TIME_FORMAT = "%Y-%m-%d-%z"
    date_cur = datetime.strptime(date_str, TIME_FORMAT)
    return float(date_cur.timestamp())


def main():
    """
    instructions for recompiling the dataset from
    https://zenodo.org/record/7323875#.ZD1-43ZKguX

    1. download all datasets into a folder specified by in_dir (such as full_dataset)
    2. run the following code to extract the needed information
    """

    # _, airports_country = load_icao_airports(fname="airport_codes.csv")

    # in_dir = "full_dataset/"
    # out_dir = "edgelists/"

    # csv_name = "flightlist_20190101_20190131.csv"

    # csv_names = find_csv_filenames(in_dir)
    # processed_lines = 0
    # skipped_lines = 0
    # miss_node_lines = 0

    # for csv_name in tqdm(csv_names):
    #     fname = in_dir + csv_name
    #     outname = out_dir + csv_name[11:-4] + "edgelist"+".csv"
    #     line_count, skip_lines, miss_node = flight2edgelist(fname, outname, node_dict=airports_country)
    #     processed_lines += line_count
    #     skipped_lines += skip_lines
    #     miss_node_lines += miss_node
    # print(f'Processed {processed_lines} lines.')
    # print(f'Skipped {skipped_lines} lines.')
    # print(f'missing node {miss_node_lines} lines.')

    """
    merge all edgelists into one file
    """
    # in_dir = "edgelists/"
    # outname = "opensky_edgelist.csv"
    # csv_names = find_csv_filenames(in_dir)
    # merge_edgelist(csv_names, in_dir, outname)

    """
    clean the node features
    """
    # in_file = "edgelists/airport_codes.csv"
    # outname = "airport_node_feat.csv"
    # clean_node_feat(in_file, outname)


    """
    sort the edgelist by day
    """
    # in_file = "tgbl-flight_edgelist.csv"
    # outname = "tgbl-flight_edgelist_sorted.csv"
    # sort_edgelist(in_file, outname)


    """
    fixing time zone different for strip time
    """
    tz_offset = "-0500"
    ts = "2021-11-29" + "-" + tz_offset
    print (date2ts(ts))


    tz_offset = "+0000"
    ts_utc = "2021-11-29" + "-" + tz_offset
    print (date2ts(ts_utc))


if __name__ == "__main__":
    main()



"""
Sample negative edges for evaluation of dynamic link prediction
Load already generated negative edges from file, batch them based on the positive edge, and return the evaluation set
"""

import torch
from torch import Tensor
import numpy as np
from torch_geometric.data import TemporalData
from tgb.utils.utils import save_pkl, load_pkl
from tgb.utils.info import PROJ_DIR
from typing import Union
import os
import time


class TKGNegativeEdgeSampler(object):
    def __init__(
        self,
        dataset_name: str,
        first_dst_id: int,
        last_dst_id: int,
        strategy: str = "time-filtered",
        partial_path: str = PROJ_DIR + "/data/processed",
    ) -> None:
        r"""
        Negative Edge Sampler
            Loads and query the negative batches based on the positive batches provided.
        constructor for the negative edge sampler class

        Parameters:
            dataset_name: name of the dataset
            first_dst_id: identity of the first destination node
            last_dst_id: indentity of the last destination node
            strategy: will always load the pre-generated negatives
            partial_path: the path to the directory where the negative edges are stored
        
        Returns:
            None
        """
        self.dataset_name = dataset_name
        self.eval_set = {}
        self.first_dst_id = first_dst_id
        self.last_dst_id = last_dst_id
        self.strategy = strategy
        self.dst_dict = None
      
    def load_eval_set(
        self,
        fname: str,
        split_mode: str = "val",
    ) -> None:
        r"""
        Load the evaluation set from disk, can be either val or test set ns samples
        Parameters:
            fname: the file name of the evaluation ns on disk
            split_mode: the split mode of the evaluation set, can be either `val` or `test`
        
        Returns:
            None
        """
        assert split_mode in [
            "val",
            "test",
        ], "Invalid split-mode! It should be `val`, `test`"
        if not os.path.exists(fname):
            raise FileNotFoundError(f"File not found at {fname}")
        self.eval_set[split_mode] = load_pkl(fname)

    def query_batch(self, 
                    pos_src: Union[Tensor, np.ndarray], 
                    pos_dst: Union[Tensor, np.ndarray], 
                    pos_timestamp: Union[Tensor, np.ndarray], 
                    edge_type: Union[Tensor, np.ndarray],
                    split_mode: str = "test") -> list:
        r"""
        For each positive edge in the `pos_batch`, return a list of negative edges
        `split_mode` specifies whether the valiation or test evaluation set should be retrieved.
        modify now to include edge type argument

        Parameters:
            pos_src: list of positive source nodes
            pos_dst: list of positive destination nodes
            pos_timestamp: list of timestamps of the positive edges
            split_mode: specifies whether to generate negative edges for 'validation' or 'test' splits

        Returns:
            neg_samples: list of numpy array; each array contains the set of negative edges that
                        should be evaluated against each positive edge.
        """
        assert split_mode in [
            "val",
            "test",
        ], "Invalid split-mode! It should be `val`, `test`!"
        if self.eval_set[split_mode] == None:
            raise ValueError(
                f"Evaluation set is None! You should load the {split_mode} evaluation set first!"
            )
        
        # check the argument types...
        if torch is not None and isinstance(pos_src, torch.Tensor):
            pos_src = pos_src.detach().cpu().numpy()
        if torch is not None and isinstance(pos_dst, torch.Tensor):
            pos_dst = pos_dst.detach().cpu().numpy()
        if torch is not None and isinstance(pos_timestamp, torch.Tensor):
            pos_timestamp = pos_timestamp.detach().cpu().numpy()
        if torch is not None and isinstance(edge_type, torch.Tensor):
            edge_type = edge_type.detach().cpu().numpy()
        
        if not isinstance(pos_src, np.ndarray) or not isinstance(pos_dst, np.ndarray) or not(pos_timestamp, np.ndarray) or not(edge_type, np.ndarray):
            raise RuntimeError(
                "pos_src, pos_dst, and pos_timestamp need to be either numpy ndarray or torch tensor!"
                )
        
        if self.strategy == "time-filtered":
            neg_samples = []
            for pos_s, pos_d, pos_t, e_type in zip(pos_src, pos_dst, pos_timestamp, edge_type):
                if (pos_t, pos_s, e_type) not in self.eval_set[split_mode]:
                    raise ValueError(
                        f"The edge ({pos_s}, {pos_d}, {pos_t}, {e_type}) is not in the '{split_mode}' evaluation set! Please check the implementation."
                    )
                else:
                    conflict_dict = self.eval_set[split_mode]
                    conflict_set = conflict_dict[(pos_t, pos_s, e_type)]
                    all_dst = np.arange(self.first_dst_id, self.last_dst_id + 1)
                    filtered_all_dst = np.delete(all_dst, conflict_set, axis=0)

                    #! always using all possible destinations for evaluation
                    neg_d_arr = filtered_all_dst

                    #! this is very slow
                    neg_samples.append(
                            neg_d_arr
                        )
        elif self.strategy == "dst-time-filtered":
            neg_samples = []
            for pos_s, pos_d, pos_t, e_type in zip(pos_src, pos_dst, pos_timestamp, edge_type):
                if (pos_t, pos_s, e_type) not in self.eval_set[split_mode]:
                    raise ValueError(
                        f"The edge ({pos_s}, {pos_d}, {pos_t}, {e_type}) is not in the '{split_mode}' evaluation set! Please check the implementation."
                    )
                else:
                    filtered_dst = self.eval_set[split_mode]
                    neg_d_arr = filtered_dst[(pos_t, pos_s, e_type)]
                    neg_samples.append(
                            neg_d_arr
                        )
        #? can't convert to numpy array due to different lengths of negative samples
        return neg_samples


        


import json
from datetime import datetime
import glob
import gzip
import csv
"""
go to https://www.gharchive.org/

wget https://data.gharchive.org/2024-01-{01..31}-{0..23}.json.gz

Creates (src, edge_type, dst, time) edges from the GitHub archive JSON file.
Using the rules from https://arxiv.org/pdf/2007.01231 (page 11)
The parser creates 18 rules that are in the GITHUB-SE-1Y-Repo dataset. I wrote the meaning of the rules and sources and destination types here.
"""



rels = {
    "IC_Created_IC_I": "IC_AO_C_I",
    "IC_Created_U_IC": "U_SO_C_IC",
    "I_Opened_U_I": "U_SE_O_I",
    "I_Opened_I_R": "I_AO_O_R",
    "I_Closed_U_I": "U_SE_C_I",
    "I_Closed_I_R": "I_AO_C_R",
    "I_Reopened_U_I": "U_SE_RO_I",
    "I_Reopened_I_R": "I_AO_RO_R",
    "PR_Opened_U_PR": "U_SO_O_P",
    "PR_Opened_PR_R": "P_AO_O_R",
    "PR_Closed_U_PR": "U_SO_C_P",
    "PR_Closed_PR_R": "P_AO_C_R",
    "PR_Reopened_U_PR": "U_SO_R_P",
    "PR_Reopened_PR_R": "P_AO_R_R",
    "PRRC_Created_U_PRC": "U_SO_C_PRC",
    "PRRC_Created_PRC_PR": "PRC_AO_C_P",
    "Forked_R_R": "R_FO_R",
    "AddMember_U_R": "U_CO_A_R",
}

issue_comment_format = "/issue_comment/{}"
issue_format = "/issue/{}"
user_format = "/user/{}"
repo_format = "/repo/{}"
pull_request_format = "/pr/{}"
pull_request_review_comment_format = "/pr_review_comment/{}"


def str_to_timestamp(time_str):
    dt = datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%SZ")
    return int(dt.timestamp())


def parse_issue_comment_events(event):
    try:
        if "action" not in event["payload"]:
            return []
        if event["payload"]["action"] == "created":
            issue_comment_id = event["payload"]["comment"]["id"]
            issue_id = event["payload"]["issue"]["id"]
            user_id = event["actor"]["id"]
            created_at = str_to_timestamp(event["created_at"])

            ici_event = [
                issue_comment_format.format(issue_comment_id),
                rels["IC_Created_IC_I"],
                issue_format.format(issue_id),
                created_at,
            ]
            uic_event = [
                user_format.format(user_id),
                rels["IC_Created_U_IC"],
                issue_comment_format.format(issue_comment_id),
                created_at,
            ]
            return [ici_event, uic_event]
        return []
    except:
        return []


def parse_issue_event(event):
    try:
        issue_id = event["payload"]["issue"]["id"]
        user_id = event["actor"]["id"]
        repo_id = event["repo"]["id"]
        created_at = str_to_timestamp(event["created_at"])
        action_map = {
            "opened": ("I_Opened_U_I", "I_Opened_I_R"),
            "closed": ("I_Closed_U_I", "I_Closed_I_R"),
            "reopened": ("I_Reopened_U_I", "I_Reopened_I_R"),
        }
        for action, event_rels in action_map.items():
            if event["payload"]["action"] == action:
                ui_event = [
                    user_format.format(user_id),
                    rels[event_rels[0]],
                    issue_format.format(issue_id),
                    created_at,
                ]

                ir_event = [
                    issue_format.format(issue_id),
                    rels[event_rels[1]],
                    repo_format.format(repo_id),
                    created_at,
                ]
                return [ui_event, ir_event]
        return []
    except:
        return []


def parse_pull_request_event(event):
    try:
        pull_request_id = event["payload"]["pull_request"]["id"]
        user_id = event["actor"]["id"]
        repo_id = event["repo"]["id"]
        created_at = str_to_timestamp(event["created_at"])
        action_map = {
            "opened": ("PR_Opened_U_PR", "PR_Opened_PR_R"),
            "closed": ("PR_Closed_U_PR", "PR_Closed_PR_R"),
            "reopened": ("PR_Reopened_U_PR", "PR_Reopened_PR_R"),
        }
        for action, event_rels in action_map.items():
            if event["payload"]["action"] == action:
                upr_event = [
                    user_format.format(user_id),
                    rels[event_rels[0]],
                    pull_request_format.format(pull_request_id),
                    created_at,
                ]

                prr_event = [
                    pull_request_format.format(pull_request_id),
                    rels[event_rels[1]],
                    repo_format.format(repo_id),
                    created_at,
                ]
                return [upr_event, prr_event]
        return []
    except:
        return []

def parse_pull_request_review_comment_event(event):
    try:
        pull_request_review_comment_id = event["payload"]["comment"]["id"]
        pull_request_id = event["payload"]["pull_request"]["id"]
        user_id = event["actor"]["id"]
        created_at = str_to_timestamp(event["created_at"])
        if event["payload"]["action"] == "created":
            uprc_event = [
                user_format.format(user_id),
                rels["PRRC_Created_U_PRC"],
                pull_request_review_comment_format.format(pull_request_review_comment_id),
                created_at,
            ]

            prcpr_event = [
                pull_request_review_comment_format.format(pull_request_review_comment_id),
                rels["PRRC_Created_PRC_PR"],
                pull_request_format.format(pull_request_id),
                created_at,
            ]
            return [uprc_event, prcpr_event]
        return []
    except:
        return []


def parse_fork_event(event):
    try:
        forkee_repo_id = event["payload"]["forkee"]["id"]
        forked_repo_id = event["repo"]["id"]
        created_at = str_to_timestamp(event["created_at"])
        return [
            [
                repo_format.format(forkee_repo_id),
                rels["Forked_R_R"],
                repo_format.format(forked_repo_id),
                created_at,
            ]
        ]
    except:
        return []


def parse_member_event(event):
    try:
        user_id = event["payload"]["member"]["id"]
        repo_id = event["repo"]["id"]
        created_at = str_to_timestamp(event["created_at"])
        return [
            [
                user_format.format(user_id),
                rels["AddMember_U_R"],
                repo_format.format(repo_id),
                created_at,
            ]
        ]
    except:
        return []


event_handler_dict = {
    "IssueCommentEvent": parse_issue_comment_events,
    "IssuesEvent": parse_issue_event,
    "PullRequestEvent": parse_pull_request_event,
    "PullRequestReviewCommentEvent": parse_pull_request_review_comment_event,
    "ForkEvent": parse_fork_event,
    "MemberEvent": parse_member_event,
}


def parse_event(event):
    event_type = event["type"]
    if event_type in event_handler_dict:
        output_list = event_handler_dict[event_type](event)
        # print("Got {} outputs for event type {}".format(len(output_list), event_type))
    else:
        # print("Unknown event type: {}".format(event_type))
        output_list = []
    return output_list


def parse_file(filename):
    # events = []
    output_dict = {}
    num_edge = 1
    #with open(filename) as f:
    with gzip.open(filename, 'r') as f:
        for i, line in enumerate(f):
            djson = json.loads(line)
            parsed_events = parse_event(djson)
            if (len(parsed_events) > 0):
                for edge in parsed_events:
                    #? ['/user/41898282', 'U_SE_O_I', '/issue/2061196208', 1704085558]
                    ts = int(edge[3])
                    head = edge[0]
                    rel = edge[1]
                    tail = edge[2]
                    if ts not in output_dict:
                        output_dict[ts] = {}
                        output_dict[ts][(head,tail,rel)] = 1
                        num_edge += 1
                    else:
                        if (head,tail,rel) in output_dict[ts]:
                            output_dict[ts][(head,tail,rel)] += 1
                        else:
                            output_dict[ts][(head,tail,rel)] = 1
                            num_edge += 1
    print("Parsed {} events".format(num_edge))
    return output_dict

def write2csv(outname, out_dict):
    with open(outname, 'a') as f:
        writer = csv.writer(f, delimiter =',')
        # writer.writerow(['ts', 'head', 'tail', 'relation_type'])
        ts_list = list(out_dict.keys())
        ts_list.sort()

        for ts in ts_list:
            for edge in out_dict[ts]:
                head = edge[0]
                tail = edge[1]
                relation_type = edge[2]
                row = [ts, head, tail, relation_type]
                writer.writerow(row)



def main():
    total_edge_dict = {}
    for file in glob.glob("*.json.gz"):
        print ("processing,", file)
        edge_dict = parse_file(file)
        # print ('check for edge overlap')
        # print(edge_dict.keys() & total_edge_dict.keys())
        # print ("-------------------------")
        #! write to csv after each file is processed. 
        # total_edge_dict.update(edge_dict)
        outname = "github_01_2024.csv"
        write2csv(outname, edge_dict)
        






if __name__ == "__main__":
    main()


import csv


def load_edgelist(file_path, freq_threshold=5):
    """
    ts, head, tail, relation_type
    1704085200,/user/34452971,/pr/1660752740,U_SO_C_P
    """
    first_row = True
    edge_dict = {}
    num_nodes = 0
    num_edges = 0
    num_rels = 0
    node_dict = {}
    edge_freq_dict = {}
    num_lines = 0


    #! identify node type with least amount of edges
    node_type_freq = {}

    with open(file_path, 'r') as f:
        reader = csv.reader(f, delimiter =',')
        for row in reader: 
            if first_row:
                first_row = False
                continue
            ts = int(row[0])
            head = row[1]
            head_type = head.split("/")[1]
            if (head_type not in node_type_freq):
                node_type_freq[head_type] = 1
            else:
                node_type_freq[head_type] += 1
            tail = row[2]
            tail_type = tail.split("/")[1]
            if (tail_type not in node_type_freq):
                node_type_freq[tail_type] = 1
            else:
                node_type_freq[tail_type] += 1
            relation_type = row[3]
            if head not in node_dict:
                node_dict[head] = 1
                num_nodes += 1
            else:
                node_dict[head] += 1

            if tail not in node_dict:
                node_dict[tail] = 1
                num_nodes += 1
            else:
                node_dict[tail] += 1

            if relation_type not in edge_freq_dict:
                edge_freq_dict[relation_type] = 1
                num_rels += 1
            else:
                edge_freq_dict[relation_type] += 1
            num_lines += 1
    print ("there are ", num_lines, " edges")
    print ("there are ", num_nodes, " nodes")
    print ("there are ", num_rels, " relations")

    node_freq5 = 0
    node_freq10 = 0
    node_freq100 = 0
    node_freq1000 = 0
    low_freq_dict = {}
    for k, v in node_dict.items():
        if v <= freq_threshold:
            low_freq_dict[k] = 1
            node_freq5 += 1
        if v >= 10:
            node_freq10 += 1
        if v >= 100:
            node_freq100 += 1
        if v >= 1000:
            node_freq1000 += 1
    print ("there are ", node_freq5, " nodes with frequency <= ", freq_threshold, " (inclusive)")
    print ("there are ", node_freq10, " nodes with frequency >= 10")
    print ("there are ", node_freq100, " nodes with frequency >= 100")
    print ("there are ", node_freq1000, " nodes with frequency >= 1000")
    # return node_freq10_dict
    return low_freq_dict, node_type_freq



def subset_by_node(file_path, low_freq_dict):
    first_row = True
    edge_dict = {}
    with open(file_path, 'r') as f:
        reader = csv.reader(f, delimiter =',')
        for row in reader: 
            if first_row:
                first_row = False
                continue
            ts = int(row[0])
            head = row[1]
            tail = row[2]
            relation_type = row[3]

            #! remove any edges that belongs any node with degree one
            if (head in low_freq_dict) or (tail in low_freq_dict):
                continue

            # if (head in node_dict) or (tail in node_dict):
            if ts not in edge_dict:
                edge_dict[ts] = {}
            if (head,tail,relation_type) not in edge_dict[ts]:
                edge_dict[ts][(head,tail,relation_type)] = 1
            else:
                edge_dict[ts][(head,tail,relation_type)] += 1
    return edge_dict


def subset_by_node_type(file_path, remove_node_type_dict, low_freq_dict=None):
    first_row = True
    edge_dict = {}
    node_dict = {}
    num_edges = 0
    if (low_freq_dict is not None):
        check_low_freq = True

    with open(file_path, 'r') as f:
        reader = csv.reader(f, delimiter =',')
        for row in reader: 
            if first_row:
                first_row = False
                continue
            ts = int(row[0])
            head = row[1]
            tail = row[2]
            if (head in low_freq_dict) or (tail in low_freq_dict):
                continue

            head_type = head.split("/")[1]
            tail_type = tail.split("/")[1]
            relation_type = row[3]



            if (head_type in remove_node_type_dict) or (tail_type in remove_node_type_dict):
                continue

            if (head not in node_dict):
                node_dict[head] = 1
            if (tail not in node_dict):
                node_dict[tail] = 1
            num_edges += 1
            if ts not in edge_dict:
                edge_dict[ts] = {}
            if (head,tail,relation_type) not in edge_dict[ts]:
                edge_dict[ts][(head,tail,relation_type)] = 1
            else:
                edge_dict[ts][(head,tail,relation_type)] += 1

    print ("there are ", num_edges, " edges in the output file")
    print ("there are ", len(node_dict), " nodes in the output file")
    return edge_dict




def write2csv(outname, out_dict):
    num_edges = 0
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['ts', 'head', 'tail', 'relation_type'])
        ts_list = list(out_dict.keys())
        ts_list.sort()

        for ts in ts_list:
            for edge in out_dict[ts]:
                head = edge[0]
                tail = edge[1]
                relation_type = edge[2]
                row = [ts, head, tail, relation_type]
                writer.writerow(row)
                num_edges += 1
    print ("there are ", num_edges, " edges in the output file")





def combine_edgelist(file_paths, outname):
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['ts', 'head', 'tail', 'relation_type'])
        for file_path in file_paths:
            first_row = True
            with open(file_path, 'r') as f:
                reader = csv.reader(f, delimiter =',')
                for row in reader: 
                    if first_row:
                        first_row = False
                        continue
                    ts = int(row[0])
                    head = row[1]
                    tail = row[2]
                    relation_type = row[3]
                    writer.writerow([ts, head, tail, relation_type])




def main():
    file_path = "github_03_2024.csv"
    freq_threshold = 2
    low_freq_dict, node_type_dict = load_edgelist(file_path, freq_threshold=freq_threshold)

    remove_node_type_dict = {'issue_comment':1, 'pr_review_comment':1} #{'issue_comment':1, 'pr_review_comment':1, 'issue':1} 
    edge_dict = subset_by_node_type(file_path, remove_node_type_dict, low_freq_dict=low_freq_dict)
    # edge_dict = subset_by_node(file_path, low_freq_dict=low_freq_dict)
    outname = "github_03_2024_subset.csv"
    write2csv(outname, edge_dict)

    # file_paths = ["github_01_2024_subset.csv", "github_02_2024_subset.csv", "github_03_2024_subset.csv"]
    # outname = "thgl-github_edges.csv"
    # combine_edgelist(file_paths, outname)





    
if __name__ == "__main__":
    main()


import time
from tgb.linkproppred.thg_negative_generator import THGNegativeEdgeGenerator
from tgb.linkproppred.dataset_pyg import PyGLinkPropPredDataset


def main():
    r"""
    Generate negative edges for the validation or test phase
    """
    print("*** Negative Sample Generation ***")

    # setting the required parameters
    num_neg_e_per_pos = 100 #-1 
    neg_sample_strategy = "node-type-filtered"
    rnd_seed = 42


    name = "thgl-forum"
    dataset = PyGLinkPropPredDataset(name=name, root="datasets")
    train_mask = dataset.train_mask
    val_mask = dataset.val_mask
    test_mask = dataset.test_mask
    data = dataset.get_TemporalData()



    data_splits = {}
    data_splits['train'] = data[train_mask]
    data_splits['val'] = data[val_mask]
    data_splits['test'] = data[test_mask]

    min_node_idx = min(int(data.src.min()), int(data.dst.min()))
    max_node_idx = max(int(data.src.max()), int(data.dst.max()))

    neg_sampler = THGNegativeEdgeGenerator(
        dataset_name=name,
        first_node_id=min_node_idx,
        last_node_id=max_node_idx,
        node_type=dataset.node_type,
        num_neg_e=num_neg_e_per_pos,
        strategy=neg_sample_strategy,
        rnd_seed=rnd_seed,
        edge_data=data,
    )

   
    # generate evaluation set
    partial_path = "."
    # generate validation negative edge set
    start_time = time.time()
    split_mode = "val"
    print(
        f"INFO: Start generating negative samples: {split_mode} --- {neg_sample_strategy}"
    )
    neg_sampler.generate_negative_samples(
        pos_edges=data_splits[split_mode], split_mode=split_mode, partial_path=partial_path
    )
    print(
        f"INFO: End of negative samples generation. Elapsed Time (s): {time.time() - start_time: .4f}"
    )

    # generate test negative edge set
    start_time = time.time()
    split_mode = "test"
    print(
        f"INFO: Start generating negative samples: {split_mode} --- {neg_sample_strategy}"
    )
    neg_sampler.generate_negative_samples(
        pos_edges=data_splits[split_mode], split_mode=split_mode, partial_path=partial_path
    )
    print(
        f"INFO: End of negative samples generation. Elapsed Time (s): {time.time() - start_time: .4f}"
    )


if __name__ == "__main__":
    main()



import csv
import datetime
import glob, os


def load_csv_raw(fname):
    r"""
    load from the raw data and retrieve, timestamp, head, tail, relation 
    convert all dates into unix timestamps
    #! Event ID	Event Date	Source Name	Source Sectors	Source Country	Event Text	CAMEO Code	Intensity	Target Name	Target Sectors	Target Country	Story ID	Sentence Number	Publisher	City	District	Province	Country	Latitude	Longitude
    """
    out_dict = {}
    first_row = True
    num_lines = 0
    with open(fname, 'r', encoding='ISO-8859-1') as f:
        reader = csv.reader(f, delimiter ='\t')
        for row in reader: 
            if first_row:
                first_row = False
                continue
            date = row[1] #1995-01-01
            head = row[2]
            tail = row[8]
            relation_type = row[6] #CAMEO code  #! not always integer in 2017 for some reason there is 13y
            if (len(date) == 0):
                continue
            
            if ("None" in date or "None" in head or "None" in tail or "None" in relation_type):
                continue
            else:
                 #! remove redundant edges with same timestamps
                TIME_FORMAT = "%Y-%m-%d" #2018-01-01
                date_cur = datetime.datetime.strptime(date, TIME_FORMAT)
                ts = int(date_cur.timestamp())
                num_lines += 1
                if (ts in out_dict):
                    if (head, tail, relation_type) in out_dict[ts]:
                        out_dict[ts][(head, tail, relation_type)] += 1
                    else:
                        out_dict[ts][(head, tail, relation_type)] = 1
                else:
                    out_dict[ts] = {}
                    out_dict[ts][(head, tail, relation_type)] = 1
    return out_dict, num_lines


def write2csv(outname, out_dict):

    node_dict = {}
    max_node_id = 0
    edge_type_dict = {}
    max_edge_type_id = 0

    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['date', 'head', 'tail', 'relation_type'])

        for date in out_dict:
            for edge in out_dict[date]:
                head = edge[0]
                tail = edge[1]
                relation_type = edge[2]
                if head not in node_dict:
                    node_dict[head] = max_node_id
                    max_node_id += 1
                if tail not in node_dict:
                    node_dict[tail] = max_node_id
                    max_node_id += 1
                if relation_type not in edge_type_dict:
                    edge_type_dict[relation_type] = max_edge_type_id
                    max_edge_type_id += 1
                row = [date, node_dict[head], node_dict[tail], edge_type_dict[relation_type]]
                writer.writerow(row)
    return node_dict, edge_type_dict


def writeEdgeTypeMapping(edge_type_dict, outname):
    r"""
    write the edge type mapping to a file
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['edge_id', 'type'])
        for key in edge_type_dict:
            writer.writerow([key, edge_type_dict[key]])




def main():

    total_lines = 0
    total_edge_dict = {} 
    #1. find all files with .txt in the folder
    for file in glob.glob("*.tab"):
        # outname = file[7:11] + "_edgelist.csv"
        print ("processing", file)
        edge_dict, num_lines = load_csv_raw(file)
        total_lines += num_lines
        print ("-----------------------------------")
        print ("file, ", file)
        print ("number of lines, ", num_lines)
        print ("number of days, ", len(edge_dict))
        print ("-----------------------------------")
        total_edge_dict.update(edge_dict)
    outname = "tkgl-icews_edgelist_tiny.csv"
    print ("total number of lines", total_lines)
    print ("total number of days", len(total_edge_dict))    
    node_dict, edge_type_dict = write2csv(outname, total_edge_dict)
    writeEdgeTypeMapping(edge_type_dict, "tkgl-icews_edgemapping.csv")




if __name__ == "__main__":
    main()


import csv

def load_static_edgelist(file_path):
    r"""
    Load the static edgelist from the file_path
    Args:
        file_path: str, The path to the file
    """
    static_dict = {}
    first_row = True
    with open(file_path, 'r') as f:
        reader = csv.reader(f, delimiter =',')
        for row in reader: 
            if first_row:
                first_row = False
                continue
            head = row[0]
            tail = row[1]
            relation_type = row[2]
            static_dict[(head, tail, relation_type)] = 1
    return static_dict


def load_temporal_edgelist(file_path):
    r"""
    Load the temporal edgelist from the file_path
    Args:
        file_path: str, The path to the file
    """
    temporal_dict = {}
    first_row = True
    with open(file_path, 'r') as f:
        """
        ts,head,tail,relation_type
        0,Q331755,Q1294765,P39
        """
        reader = csv.reader(f, delimiter =',')
        for row in reader: 
            if first_row:
                first_row = False
                continue
            ts = int(row[0])
            head = row[1]
            tail = row[2]
            relation_type = row[3]
            if ts not in temporal_dict:
                temporal_dict[ts] = {}
                temporal_dict[ts][(head, tail, relation_type)] = 1
            else:
                if (head, tail, relation_type) in temporal_dict[ts]:
                    temporal_dict[ts][(head, tail, relation_type)] += 1
                else:
                    temporal_dict[ts][(head, tail, relation_type)] = 1
    return temporal_dict


def remove_conflict(static_dict, temporal_dict):
    r"""
    Remove the conflict between the static and temporal edgelist
    Args:
        static_dict: dict, The static edgelist
        temporal_dict: dict, The temporal edgelist
    """
    num_conflicts = 0
    for ts in temporal_dict:
        for edge in temporal_dict[ts]:
            head = edge[0]
            tail = edge[1]
            relation_type = edge[2]
            if (head, tail, relation_type) in static_dict:
                num_conflicts += 1
                static_dict.pop((head, tail, relation_type))
    print("Removed {} conflicts".format(num_conflicts))
    return static_dict


def write2csv(outname: str, 
              out_dict: dict,):
    r"""
    Write the dictionary to a csv file
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        #head,tail,relation_type
        writer.writerow(['head', 'tail', 'relation_type'])
        for edge in out_dict:
            head = edge[0]
            tail = edge[1]
            relation_type = edge[2]
            row = [head, tail, relation_type]
            writer.writerow(row)




def main():
    #! remove conflict: remove all edges with the same head, tail, relation_type from the static edgelist
    static_file = "tkgl-smallpedia_static_edgelist.csv"
    temporal_file = "tkgl-smallpedia_edgelist.csv"
    static_dict = load_static_edgelist(static_file)
    print("constructed static dictionary")
    temporal_dict = load_temporal_edgelist(temporal_file)
    print("constructed temporal dictionary")
    static_dict = remove_conflict(static_dict, temporal_dict)
    out_name = "tkgl-smallpedia_static_edgelist_no_conflict.csv"
    write2csv(out_name, static_dict)




if __name__ == "__main__":
    main()


import csv
import datetime
import glob, os


def load_time_csv(fname):
    r"""
    load from data and retrieve, ts,head,tail,relation_type,time_rel_type
    """
    out_dict = {} #only contain edges {ts: {(head, tail, rel_type):count}}
    start_end_dict = {} #{(head, tail, rel_type): {start:year, end:year}}

    first_row = True
    point_in_time_lines = 0
    start_end_lines = 0

    #? ts,head,tail,relation_type,time_rel_type
    #* 0,Q331755,Q1294765,P39,P580
    with open(fname, 'r') as f:
        reader = csv.reader(f, delimiter =',')
        for row in reader: 
            if first_row:
                first_row = False
                continue
            ts = int(row[0])
            head = row[1]
            tail = row[2]
            relation_type = row[3]
            time_rel = row[4]
            if (time_rel in ['P585', 'P577', 'P574']):
                if (ts in out_dict):
                    if (head, tail, relation_type) in out_dict[ts]:
                        out_dict[ts][(head, tail, relation_type)] += 1
                    else:
                        out_dict[ts][(head, tail, relation_type)] = 1
                else:
                    out_dict[ts] = {}
                    out_dict[ts][(head, tail, relation_type)] = 1
                point_in_time_lines += 1
            else: # time_rel in ['P580', 'P582']
                if (head, tail, relation_type) in start_end_dict:
                    if (time_rel in ['P580']):
                        start_end_dict[(head, tail, relation_type)]['start'] = ts
                    elif (time_rel in ['P582']):
                        start_end_dict[(head, tail, relation_type)]['end'] = ts
                    else:
                        raise ValueError(f"Unknown time_rel: {time_rel}")
                else:
                    start_end_dict[(head, tail, relation_type)] = {}
                    if (time_rel in ['P580']):
                        start_end_dict[(head, tail, relation_type)]['start'] = ts
                    else:
                        start_end_dict[(head, tail, relation_type)]['end'] = ts
                start_end_lines += 1

    print ("-----------------------------------")
    print ("for this edgelist:")
    print (f"point_in_time_lines: {point_in_time_lines}")
    print (f"start_end_lines: {start_end_lines}")
    print ("-----------------------------------")
    
    

    repeated_lines = 0
    no_duration_lines = 0
    
    #* now, repeat edges from start_end_dict
    for edge in start_end_dict.keys():
        if 'start' not in start_end_dict[edge]:
            #start_end_dict[edge]['start'] = 0 #start at year 0
            #start_end_dict[edge]['start'] = start_end_dict[edge]['end']
            no_duration_lines += 1
            continue
        if 'end' not in start_end_dict[edge]:
            # start_end_dict[edge]['end'] = 2024 #end at year 2024
            start_end_dict[edge]['end'] = start_end_dict[edge]['start'] #end at year 2024
            no_duration_lines += 1
            continue
        for year in range(start_end_dict[edge]['start'], start_end_dict[edge]['end']+1):
            if year not in out_dict:
                out_dict[year] = {}
            out_dict[year][edge] = 1
            repeated_lines += 1

    print ("-----------------------------------")
    print ("for this edgelist:")
    print (f"point_in_time_lines: {point_in_time_lines}")
    print (f"start_end_lines: {start_end_lines} resulting in")
    print (f"repeated_lines: {repeated_lines}")
    print (f"no_duration_lines: {no_duration_lines}")
    print ("-----------------------------------")
    print ("total lines: ", point_in_time_lines + repeated_lines)
    num_lines = point_in_time_lines + repeated_lines
    return out_dict, num_lines



def write2csv(outname: str, 
              out_dict: dict,):
    r"""
    Write the dictionary to a csv file
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['ts', 'head', 'tail', 'relation_type'])
        dates = list(out_dict.keys())
        dates.sort()
        for date in dates:
            for edge in out_dict[date]:
                head = edge[0]
                tail = edge[1]
                relation_type = edge[2]
                row = [date, head, tail, relation_type]
                writer.writerow(row)


def extract_subset(fname, outname, start_year=2000, end_year=2024):
    node_dict = {}
    first_row = True
    rel_type = {}
    r"""
    ts,head,tail,relation_type
    0,Q331755,Q1294765,P39
    0,Q116233388,Q2566630,P2348
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['ts', 'head', 'tail', 'relation_type'])
        with open(fname, 'r') as f:
            reader = csv.reader(f, delimiter =',')
            for row in reader: 
                if first_row:
                    first_row = False
                    continue
                ts = int(row[0])
                head = row[1]
                tail = row[2]
                relation_type = row[3]
                if (ts >= start_year and ts <= end_year):
                    if head not in node_dict:
                        node_dict[head] = 1
                    if tail not in node_dict:
                        node_dict[tail] = 1
                    row = [ts, head, tail, relation_type]
                    if (relation_type not in rel_type):
                        rel_type[relation_type] = 1
                    writer.writerow(row)
    print ("there are ",len(rel_type), " relation types")
    return node_dict


def extract_subset_nodeid(fname, outname, start_year=2000, end_year=2024, max_id=1000000):
    node_dict = {}
    first_row = True
    rel_type = {}
    r"""
    ts,head,tail,relation_type
    0,Q331755,Q1294765,P39
    0,Q116233388,Q2566630,P2348
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['ts', 'head', 'tail', 'relation_type'])
        with open(fname, 'r') as f:
            reader = csv.reader(f, delimiter =',')
            for row in reader: 
                if first_row:
                    first_row = False
                    continue
                ts = int(row[0])
                head = row[1]
                head_id = int(head[1:])
                tail = row[2]
                tail_id = int(tail[1:])
                if (head_id > max_id or tail_id > max_id):
                    continue
                relation_type = row[3]
                if (ts >= start_year and ts <= end_year):
                    if head not in node_dict:
                        node_dict[head] = 1
                    if tail not in node_dict:
                        node_dict[tail] = 1
                    row = [ts, head, tail, relation_type]
                    if (relation_type not in rel_type):
                        rel_type[relation_type] = 1
                    writer.writerow(row)
    print ("there are ",len(rel_type), " relation types")
    return node_dict




def extract_static_subset(fname, outname, node_dict, max_id=1000000):
    r"""
    extract static edges based a given node dict
    """
    first_row = True
    r"""
    head,tail,relation_type
    Q31,Q1088364,P1344
    Q31,Q3247091,P1151
    """
    rel_type = {}
    full_node = {}
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['head', 'tail', 'relation_type'])
        with open(fname, 'r') as f:
            reader = csv.reader(f, delimiter =',')
            for row in reader: 
                if first_row:
                    first_row = False
                    continue
                head = row[0]
                head_id = int(head[1:])
                tail = row[1]
                tail_id = int(tail[1:])
                relation_type = row[2]
                if (head_id > max_id or tail_id > max_id):
                    continue
                if (head in node_dict) or (tail in node_dict): #need to check
                    row = [head, tail, relation_type]
                    writer.writerow(row)
                    if (relation_type not in rel_type):
                        rel_type[relation_type] = 1
                    else:
                        rel_type[relation_type] += 1
                    if (head not in full_node):
                        full_node[head] = 1
                    if (tail not in full_node):
                        full_node[tail] = 1
    print ("there are ",len(rel_type), " relation types")
    print ("there are ",len(full_node), " nodes in static edgelist")
    return rel_type



#! not used, filter by top edgetypes 
def subset_static_edges(fname, outname, rel_type, topk=10):
    #* select edges based on frequency
    import operator
    sorted_x = sorted(rel_type.items(), key=operator.itemgetter(1))
    sorted_x = sorted_x[-topk:]
    rel_kept = {}
    for (u,v) in sorted_x:
        rel_kept[u] = 1
        print (u,v)

    kept_nodes = {}
    first_row = True

    # rel_kept = {"P17":1, "P27":1, "P495":1, "P19": 1}
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['head', 'tail', 'relation_type'])
        with open(fname, 'r') as f:
            reader = csv.reader(f, delimiter =',')
            for row in reader: 
                if first_row:
                    first_row = False
                    continue
                head = row[0]
                tail = row[1]
                relation_type = row[2]
                if (relation_type in rel_kept):
                    row = [head, tail, relation_type]
                    if (head not in kept_nodes):
                        kept_nodes[head] = 1
                    if (tail not in kept_nodes):
                        kept_nodes[tail] = 1
                    writer.writerow(row)
    print ("there are ",len(kept_nodes), " nodes in static edgelist")
                






def main():

    # #* repeat the edges of start and end dates
    # """
    # P580: start time
    # P582: end time
    # P585: point in time
    # P577: publication date
    # P574: year of publication of scientific name for taxon

    # we need to:
    # 1. get all edges with P585, P577 and P574 
    # 2. find out which edges has both start and end time
    # 3. for those without start time, start at year 0, without end time, end at year 2024
    # """
    # fname = "tkgl-wikidata_edgelist_raw.csv"
    # out_dict, num_lines = load_time_csv(fname)

    # outname = "tkgl-wikidata_edgelist.csv"
    # write2csv(outname, out_dict)

    inputfile = "tkgl-wikidata_edgelist.csv"
    outname = "tkgl-smallpedia_edgelist.csv"
    # start_year = 2015
    start_year=1900#1700
    end_year=2024#1800
    max_id=1000000
    # node_dict = extract_subset(inputfile, outname, start_year=start_year, end_year=end_year)
    node_dict = extract_subset_nodeid(inputfile, outname, start_year=start_year, end_year=end_year, max_id=max_id)
    print ("there are ",len(node_dict), " nodes")

    inputfile = "tkgl-wikidata_static_edgelist.csv"
    outname = "tkgl-smallpedia_static_edgelist.csv"
    rel_type = extract_static_subset(inputfile, outname, node_dict, max_id=max_id)

    #! not used
    # inputfile = "tkgl-smallpedia_static_edgelist.csv"
    # outname = "tkgl-smallpedia_static_edgelist_top10.csv"
    # topk=10
    # subset_static_edges(inputfile, outname, rel_type, topk=topk)
    





if __name__ == "__main__":
    main()


import csv
import datetime
import glob, os

def load_csv_raw(fname):
    """
    load the raw csv file and merge them into one
    """
    out_dict = {}
    num_lines = 0
    with open(fname, 'r') as f:
        reader = csv.reader(f, delimiter ='\t')
        #* /user/10746682	U_SO_C_IC	/issue_comment/455195715	1547754198
        for row in reader: 
            head = row[0]
            relation_type = row[1]
            tail = row[2]
            ts = int(row[3])
            if (ts in out_dict):
                if (head, tail, relation_type) in out_dict[ts]:
                    out_dict[ts][(head, tail, relation_type)] += 1
                else:
                    out_dict[ts][(head, tail, relation_type)] = 1
            else:
                out_dict[ts] = {}
                out_dict[ts][(head, tail, relation_type)] = 1
            num_lines += 1
    return out_dict, num_lines


def write2csv(outname, out_dict):
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['ts', 'head', 'tail', 'relation_type'])
        ts_list = list(out_dict.keys())
        ts_list.sort()

        for ts in ts_list:
            for edge in out_dict[ts]:
                head = edge[0]
                tail = edge[1]
                relation_type = edge[2]
                row = [ts, head, tail, relation_type]
                writer.writerow(row)


def load_edgelist(fname):
    """
    load the edgelist
    """
    node_dict = {} # {node_name: node_id}
    node_type_dict = {} # {node_id: node_type}
    rel_type_dict = {}
    edge_dict = {} # {edge: edge_type}
    node_type_mapping = {}
    num_edges = 0
    with open(fname, 'r') as f:
        reader = csv.reader(f, delimiter =',')
        first_row = True
        for row in reader:
            if first_row:
                first_row = False
                continue
            ts = int(row[0])
            head = row[1]
            tail = row[2]
            relation_type = row[3]

            head_strs = head.split('/')
            tail_strs = tail.split('/')
            head_type = head_strs[1]
            tail_type = tail_strs[1]

            if head_type not in node_type_mapping:
                node_type_mapping[head_type] = len(node_type_mapping)
            if tail_type not in node_type_mapping:
                node_type_mapping[tail_type] = len(node_type_mapping)

            if head not in node_dict:
                node_dict[head] = len(node_dict)
                node_type_dict[node_dict[head]] = node_type_mapping[head_type]
            if tail not in node_dict:
                node_dict[tail] = len(node_dict)
                node_type_dict[node_dict[tail]] = node_type_mapping[tail_type]
            if relation_type not in rel_type_dict:
                rel_type_dict[relation_type] = len(rel_type_dict)
            if ts not in edge_dict:
                edge_dict[ts] = {}
            edge_dict[ts][(node_dict[head], node_dict[tail], rel_type_dict[relation_type])] = 1
            num_edges += 1
    print ("there are {} nodes".format(len(node_dict)))
    print ("there are {} edges".format(num_edges))

    return node_dict, node_type_dict, edge_dict, rel_type_dict, node_type_mapping



def writeNodeType(node_type_dict, outname):
    r"""
    write the node type mapping to a file
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['node_id', 'type'])
        for key in node_type_dict:
            writer.writerow([key, node_type_dict[key]])


def writeEdgeTypeMapping(edge_type_dict, outname):
    r"""
    write the edge type mapping to a file
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['edge_id', 'type'])
        for key in edge_type_dict:
            writer.writerow([key, edge_type_dict[key]])


def writeNodeTypeMapping(node_type_dict, outname):
    r"""
    write the edge type mapping to a file
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['node_type_id', 'type'])
        for key in node_type_dict:
            writer.writerow([key, node_type_dict[key]])


def write2edgelist(out_dict, outname):
    r"""
    Write the dictionary to a csv file
    """
    num_lines = 0
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['timestamp', 'head', 'tail', 'relation_type'])
        dates = list(out_dict.keys())
        dates.sort()
        for date in dates:
            for edge in out_dict[date]:
                head = edge[0]
                tail = edge[1]
                relation_type = int(edge[2])
                row = [date, head, tail, relation_type]
                writer.writerow(row)
                num_lines += 1
    print ("there are {} lines in the file".format(num_lines))



def main():
    # """
    # concatenate edgelists
    # """
    # total_lines = 0
    # total_edge_dict = {} 
    # #1. find all files with .txt in the folder
    # for file in glob.glob("*.txt"):
    #     # outname = file[7:11] + "_edgelist.csv"
    #     print ("processing", file)
    #     edge_dict, num_lines = load_csv_raw(file)
    #     total_lines += num_lines
    #     print ("-----------------------------------")
    #     print ("file, ", file)
    #     print ("number of lines, ", num_lines)
    #     print ("number of ts, ", len(edge_dict))
    #     print ("-----------------------------------")
    #     total_edge_dict.update(edge_dict)
    # outname = "all_edgelist.csv"
    # write2csv(outname, total_edge_dict)


    fname ="github_03_2024_subset.csv"#"github_01_2024_subset.csv" #"thgl-github_edges.csv" #"all_edgelist.csv" 
    node_dict, node_type_dict, edge_dict, edge_type_dict, node_type_mapping = load_edgelist(fname)
    write2edgelist (edge_dict, "thgl-github_edgelist.csv")
    writeNodeType(node_type_dict, "thgl-github_nodetype.csv")
    writeEdgeTypeMapping(edge_type_dict, "thgl-github_edgemapping.csv")
    writeNodeTypeMapping(node_type_mapping, "thgl-github_nodemapping.csv")
    


if __name__ == "__main__":
    main()


import json
from datetime import datetime

rels = {
    "IC_Created_IC_I": "IC_AO_C_I",
    "IC_Created_U_IC": "U_SO_C_IC",
    "I_Opened_U_I": "U_SE_O_I",
    "I_Opened_I_R": "I_AO_O_R",
    "I_Closed_U_I": "U_SE_C_I",
    "I_Closed_I_R": "I_AO_C_R",
    "I_Reopened_U_I": "U_SE_RO_I",
    "I_Reopened_I_R": "I_AO_RO_R",
    "PR_Opened_U_PR": "U_SO_O_P",
    "PR_Opened_PR_R": "P_AO_O_R",
    "PR_Closed_U_PR": "U_SO_C_P",
    "PR_Closed_PR_R": "P_AO_C_R",
    "PR_Reopened_U_PR": "U_SO_R_P",
    "PR_Reopened_PR_R": "P_AO_R_R",
    "PRRC_Created_U_PRC": "U_SO_C_PRC",
    "PRRC_Created_PRC_PR": "PRC_AO_C_P",
    "Forked_R_R": "R_FO_R",
    "AddMember_U_R": "U_CO_A_R",
}

issue_comment_format = "/issue_comment/{}"
issue_format = "/issue/{}"
user_format = "/user/{}"
repo_format = "/repo/{}"
pull_request_format = "/pr/{}"
pull_request_review_comment_format = "/pr_review_comment/{}"


def str_to_timestamp(time_str):
    dt = datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%SZ")
    return int(dt.timestamp())


def parse_issue_comment_events(event):
    if event["payload"]["action"] == "created":
        issue_comment_id = event["payload"]["comment"]["id"]
        issue_id = event["payload"]["issue"]["id"]
        user_id = event["actor"]["id"]
        created_at = str_to_timestamp(event["created_at"])

        ici_event = [
            issue_comment_format.format(issue_comment_id),
            rels["IC_Created_IC_I"],
            issue_format.format(issue_id),
            created_at,
        ]
        uic_event = [
            user_format.format(user_id),
            rels["IC_Created_U_IC"],
            issue_comment_format.format(issue_comment_id),
            created_at,
        ]
        return [ici_event, uic_event]
    return []


def parse_issue_event(event):
    issue_id = event["payload"]["issue"]["id"]
    user_id = event["actor"]["id"]
    repo_id = event["repo"]["id"]
    created_at = str_to_timestamp(event["created_at"])
    action_map = {
        "opened": ("I_Opened_U_I", "I_Opened_I_R"),
        "closed": ("I_Closed_U_I", "I_Closed_I_R"),
        "reopened": ("I_Reopened_U_I", "I_Reopened_I_R"),
    }
    for action, event_rels in action_map.items():
        if event["payload"]["action"] == action:
            ui_event = [
                user_format.format(user_id),
                rels[event_rels[0]],
                issue_format.format(issue_id),
                created_at,
            ]

            ir_event = [
                issue_format.format(issue_id),
                rels[event_rels[1]],
                repo_format.format(repo_id),
                created_at,
            ]
            return [ui_event, ir_event]
    return []


def parse_pull_request_event(event):
    pull_request_id = event["payload"]["pull_request"]["id"]
    user_id = event["actor"]["id"]
    repo_id = event["repo"]["id"]
    created_at = str_to_timestamp(event["created_at"])
    action_map = {
        "opened": ("PR_Opened_U_PR", "PR_Opened_PR_R"),
        "closed": ("PR_Closed_U_PR", "PR_Closed_PR_R"),
        "reopened": ("PR_Reopened_U_PR", "PR_Reopened_PR_R"),
    }
    for action, event_rels in action_map.items():
        if event["payload"]["action"] == action:
            upr_event = [
                user_format.format(user_id),
                rels[event_rels[0]],
                pull_request_format.format(pull_request_id),
                created_at,
            ]

            prr_event = [
                pull_request_format.format(pull_request_id),
                rels[event_rels[1]],
                repo_format.format(repo_id),
                created_at,
            ]
            return [upr_event, prr_event]
    return []


def parse_pull_request_review_comment_event(event):
    pull_request_review_comment_id = event["payload"]["comment"]["id"]
    pull_request_id = event["payload"]["pull_request"]["id"]
    user_id = event["actor"]["id"]
    created_at = str_to_timestamp(event["created_at"])
    if event["payload"]["action"] == "created":
        uprc_event = [
            user_format.format(user_id),
            rels["PRRC_Created_U_PRC"],
            pull_request_review_comment_format.format(pull_request_review_comment_id),
            created_at,
        ]

        prcpr_event = [
            pull_request_review_comment_format.format(pull_request_review_comment_id),
            rels["PRRC_Created_PRC_PR"],
            pull_request_format.format(pull_request_id),
            created_at,
        ]
        return [uprc_event, prcpr_event]
    return []


def parse_fork_event(event):
    forkee_repo_id = event["payload"]["forkee"]["id"]
    forked_repo_id = event["repo"]["id"]
    created_at = str_to_timestamp(event["created_at"])
    return [
        [
            repo_format.format(forkee_repo_id),
            rels["Forked_R_R"],
            repo_format.format(forked_repo_id),
            created_at,
        ]
    ]


def parse_member_event(event):
    user_id = event["payload"]["member"]["id"]
    repo_id = event["repo"]["id"]
    created_at = str_to_timestamp(event["created_at"])
    return [
        [
            user_format.format(user_id),
            rels["AddMember_U_R"],
            repo_format.format(repo_id),
            created_at,
        ]
    ]


event_handler_dict = {
    "IssueCommentEvent": parse_issue_comment_events,
    "IssuesEvent": parse_issue_event,
    "PullRequestEvent": parse_pull_request_event,
    "PullRequestReviewCommentEvent": parse_pull_request_review_comment_event,
    "ForkEvent": parse_fork_event,
    "MemberEvent": parse_member_event,
}


def parse_event(event):
    event_type = event["type"]
    if event_type in event_handler_dict:
        output_list = event_handler_dict[event_type](event)
        # print("Got {} outputs for event type {}".format(len(output_list), event_type))
    else:
        # print("Unknown event type: {}".format(event_type))
        output_list = []
    return output_list


def parse_file(filename):
    events = []
    with open(filename) as f:
        for i, line in enumerate(f):
            event = json.loads(line)
            parsed_events = parse_event(event)
            events.append(parsed_events)
    events = [event for sublist in events for event in sublist]
    print("Parsed {} events".format(len(events)))
    return events


filename = "2015-01-01-15.json"
parse_file(filename)



import csv

"""
#! analyze statistics from the dataset
#* 1). # of unique nodes, 2). # of edges. 3). # of unique edges, 4). # of timestamps 5). min & max of edge weights, 6). recurrence of nodes
"""


def analyze_csv(fname):
    node_dict = {}
    edge_dict = {}
    num_edges = 0
    num_time = 0
    prev_t = "none"
    min_w = 100000
    max_w = 0

    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        line_count = 0
        for row in csv_reader:
            if line_count == 0:
                line_count += 1
            else:
                # t,u,v,w
                t = row[0]
                u = row[1]
                v = row[2]
                w = float(row[3].strip())

                # min & max edge weights
                if w > max_w:
                    max_w = w

                if w < min_w:
                    min_w = w

                # count unique time
                if t != prev_t:
                    num_time += 1
                    prev_t = t

                # unique nodes
                if u not in node_dict:
                    node_dict[u] = 1
                else:
                    node_dict[u] += 1

                if v not in node_dict:
                    node_dict[v] = 1
                else:
                    node_dict[v] += 1

                # unique edges
                num_edges += 1
                if (u, v) not in edge_dict:
                    edge_dict[(u, v)] = 1
                else:
                    edge_dict[(u, v)] += 1

    print("----------------------high level statistics-------------------------")
    print("number of total edges are ", num_edges)
    print("number of nodes are ", len(node_dict))
    print("number of unique edges are ", len(edge_dict))
    print("number of unique timestamps are ", num_time)
    print("maximum edge weight is ", max_w)
    print("minimum edge weight is ", min_w)

    num_10 = 0
    num_100 = 0
    num_1000 = 0

    for node in node_dict:
        if node_dict[node] >= 10:
            num_10 += 1
        if node_dict[node] >= 100:
            num_100 += 1
        if node_dict[node] >= 1000:
            num_1000 += 1
    print("number of nodes with # edges >= 10 is ", num_10)
    print("number of nodes with # edges >= 100 is ", num_100)
    print("number of nodes with # edges >= 1000 is ", num_1000)
    print("----------------------high level statistics-------------------------")


"""
return a node dict only keeping nodes with > 10 edges
"""


def extract_node_dict(fname, freq=10):
    node_dict = {}
    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        line_count = 0
        for row in csv_reader:
            if line_count == 0:
                line_count += 1
            else:
                # t,u,v,w
                t = row[0]
                u = row[1]
                v = row[2]
                w = float(row[3].strip())
                if u not in node_dict:
                    node_dict[u] = 1
                else:
                    node_dict[u] += 1

                if v not in node_dict:
                    node_dict[v] = 1
                else:
                    node_dict[v] += 1

    out_dict = {}
    for node in node_dict:
        if node_dict[node] >= freq:
            out_dict[node] = node_dict[node]
    return out_dict


"""
remove any edges do not contain either src or dst not in the node dict
"""


def clean_edgelist(fname, outname, node_dict):
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["time", "src", "dst", "weight"]
        write.writerow(fields)
        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            for row in csv_reader:
                if line_count == 0:
                    line_count += 1
                else:
                    # t,u,v,w
                    t = row[0]
                    u = row[1]
                    v = row[2]
                    w = float(row[3].strip())
                    if u in node_dict and v in node_dict:
                        write.writerow([t, u, v, w])



def sort_edgelist(in_file, outname):
    """
    sort the edges by timestamp
    """
    row_dict = {} #{day: {row: row}}
    line_idx = 0
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["day", "src", "dst", "callsign", "typecode"]
        write.writerow(fields)
        with open(in_file, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            for row in csv_reader:
                if line_idx == 0:  # header
                    line_idx += 1
                    continue
                ts = int(row[0])
                if ts not in row_dict:
                    row_dict[ts] = {}
                    row_dict[ts][line_idx] = row
                else:
                    row_dict[ts][line_idx] = row
                line_idx += 1
        
        for ts in sorted(row_dict.keys()):
            for idx in row_dict[ts].keys():
                row = row_dict[ts][idx]
                write.writerow(row)




def main():
    """
    keeping subgraph of most active nodes
    """
    # freq = 10
    # fname = "stablecoin_edgelist.csv"
    # node_dict = extract_node_dict(fname, freq=freq)

    # outname = "stablecoin_freq10.csv"
    # clean_edgelist(fname, outname, node_dict)

    # fname = "stablecoin_freq10.csv"
    # analyze_csv(fname)

    """
    sort edgelist by time
    """
    in_file = "tgbl-coin_edgelist.csv"
    outname = "tgbl-coin_edgelist_sorted.csv"
    sort_edgelist(in_file, outname)


if __name__ == "__main__":
    main()



"""
script for generating statistics from the dataset
"""
import csv
import numpy as np
# import matplotlib.pyplot as plt


"""
#! analyze statistics from the dataset
#* 1). # of unique nodes, 2). # of edges. 3). # of unique edges, 4). # of timestamps 5). recurrence of nodes
"""


def analyze_csv(fname):
    node_dict = {}
    edge_dict = {}
    num_edges = 0
    num_time = 0
    time_dict = {}

    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        line_count = 0
        for row in csv_reader:
            if line_count == 0:
                line_count += 1
            else:
                # t,u,v,w
                t = row[0]
                u = row[1]
                v = row[2]

                # count unique time
                if t not in time_dict:
                    time_dict[t] = 1
                    num_time += 1

                # unique nodes
                if u not in node_dict:
                    node_dict[u] = 1
                else:
                    node_dict[u] += 1

                if v not in node_dict:
                    node_dict[v] = 1
                else:
                    node_dict[v] += 1

                # unique edges
                num_edges += 1
                if (u, v) not in edge_dict:
                    edge_dict[(u, v)] = 1
                else:
                    edge_dict[(u, v)] += 1

    print("----------------------high level statistics-------------------------")
    print("number of total edges are ", num_edges)
    print("number of nodes are ", len(node_dict))
    print("number of unique edges are ", len(edge_dict))
    print("number of unique timestamps are ", num_time)

    num_10 = 0
    num_100 = 0
    num_1000 = 0

    for node in node_dict:
        if node_dict[node] >= 10:
            num_10 += 1
        if node_dict[node] >= 100:
            num_100 += 1
        if node_dict[node] >= 1000:
            num_1000 += 1
    print("number of nodes with # edges >= 10 is ", num_10)
    print("number of nodes with # edges >= 100 is ", num_100)
    print("number of nodes with # edges >= 1000 is ", num_1000)
    print("----------------------high level statistics-------------------------")


def plot_curve(y: np.ndarray, outname: str) -> None:
    """
    plot the training curve given y
    Parameters:
        y: np.ndarray, the training curve
        outname: str, the output name
    """
    plt.plot(y, color="#fc4e2a")
    plt.savefig(outname + ".pdf")
    plt.close()


def main():
    fname = "tgb/datasets/tgbl-wiki/tgbl-wiki_edgelist.csv"
    analyze_csv(fname)


if __name__ == "__main__":
    main()



import networkx as nx
import numpy as np
import matplotlib.pyplot as plt
import csv
from typing import Optional, Dict, Any, Tuple
import datetime
from datetime import date, timedelta
from difflib import SequenceMatcher


# similarity_dict = {('electronic', 'electronica'): 0.9523809523809523, ('electronic', 'electro'): 0.8235294117647058, ('alternative', 'alternative rock'): 0.8148148148148148, ('nu jazz', 'nu-jazz'): 0.8571428571428571,
#                    ('funky', 'funk'): 0.8888888888888888, ('funky', 'funny'): 0.8, ('post rock', 'pop rock'): 0.8235294117647058, ('post rock', 'post-rock'): 0.8888888888888888,
#                    ('instrumental', 'instrumental rock'): 0.8275862068965517, ('chill', 'chile'): 0.8, ('Drum and bass', 'Drum n Bass'): 0.8333333333333334, ('female vocalists', 'female vocalist'): 0.967741935483871,
#                    ('female vocalists', 'male vocalists'): 0.9333333333333333, ('female vocalists', 'male vocalist'): 0.896551724137931, ('electro', 'electropop'): 0.8235294117647058, ('funk', 'fun'): 0.8571428571428571,
#                    ('hip hop', 'trip hop'): 0.8, ('hip hop', 'hiphop'): 0.9230769230769231, ('trip-hop', 'trip hop'): 0.875, ('indie rock', 'indie folk'): 0.8, ('new age', 'new wave'): 0.8, ('new age', 'new rave'): 0.8,
#                    ('synthpop', 'synth pop'): 0.9411764705882353, ('industrial', 'industrial rock'): 0.8, ('cover', 'covers'): 0.9090909090909091, ('post hardcore', 'post-hardcore'): 0.9230769230769231, ('mathcore', 'deathcore'): 0.8235294117647058,
#                    ('deutsch', 'dutch'): 0.8333333333333334, ('swing', 'sting'): 0.8, ('female vocalist', 'male vocalists'): 0.896551724137931, ('female vocalist', 'male vocalist'): 0.9285714285714286, ('new wave', 'new rave'): 0.875,
#                    ('male vocalists', 'male vocalist'): 0.9629629629629629, ('Progressive rock', 'Progressive'): 0.8148148148148148, ('Alt-country', 'alt country'): 0.8181818181818182, ('favorites', 'Favourites'): 0.8421052631578947,
#                    ('favorites', 'favourite'): 0.8888888888888888, ('favorites', 'Favorite'): 0.8235294117647058, ('1970s', '1980s'): 0.8, ('1970s', '1990s'): 0.8, ('proto-punk', 'post-punk'): 0.8421052631578947,
#                    ('folk rock', 'folk-rock'): 0.8888888888888888, ('1980s', '1990s'): 0.8, ('favorite songs', 'Favourite Songs'): 0.8275862068965517, ('melancholic', 'melancholy'): 0.8571428571428571,
#                    ('Favourites', 'favourite'): 0.8421052631578947, ('Favourites', 'Favorite'): 0.8888888888888888, ('Favourites', 'Favourite Songs'): 0.8, ('favourite', 'Favorite'): 0.8235294117647058,
#                    ('american', 'americana'): 0.9411764705882353, ('american', 'african'): 0.8, ('american', 'mexican'): 0.8, ('rock en español', 'Rock en Espanol'): 0.8, ('trance', 'psytrance'): 0.8,
#                    ('power pop', 'powerpop'): 0.9411764705882353, ('psychill', 'psychobilly'): 0.8421052631578947, ('Progressive metal', 'progressive death metal'): 0.8, ('Progressive metal', 'progressive black metal'): 0.8,
#                    ('progressive death metal', 'progressive black metal'): 0.8260869565217391, ('romantic', 'new romantic'): 0.8, ('hair metal', 'Dark metal'): 0.8, ('melodic metal', 'melodic black metal'): 0.8125,
#                    ('funk metal', 'folk metal'): 0.8, ('death metal', 'math metal'): 0.8571428571428571, ('Technical Metal', 'Technical Death Metal'): 0.8333333333333334, ('speed metal', 'sid metal'): 0.8}

#! map diferent spelling and similar ones to the same one, use space if possible
# ? key = to replace, value = to keep

similarity_dict = {
    "nu-jazz": "nu jazz",
    "funky": "funk",
    "post-rock": "post rock",
    "Drum n Bass": "Drum and bass",
    "female vocalists": "female vocalist",
    "male vocalists": "male vocalist",
    "hiphop": "hip hop",
    "trip-hop": "trip hop",
    "synthpop": "synth pop",
    "covers": "cover",
    "post-hardcore": "post hardcore",
    "Favourites": "favorites",
    "favourite": "favorites",
    "Favorite": "favorites",
    "folk-rock": "folk rock",
    "favorite songs": "favorites",
    "Favourite Songs": "favorites",
    "americana": "american",
    "Rock en Espanol": "rock en español",
    "melancholy": "melancholic",
    "powerpop": "power pop",
}


def filter_genre_edgelist(fname, genres_dict):
    """
    rewrite the edgelist but only keeping the genres with high frequency, also uses similarity_dict
    """
    edgelist = open(fname, "r")
    lines = list(edgelist.readlines())
    edgelist.close()

    with open("lastfm_edgelist_clean.csv", "w") as f:
        write = csv.writer(f)
        fields = ["user_id", "timestamp", "tags", "weight"]
        write.writerow(fields)

        for i in range(1, len(lines)):
            vals = lines[i].split(",")
            user_id = vals[1]
            time = vals[2]
            genre = vals[3].strip('"').strip("['")
            w = vals[4][:-3]
            if genre in genres_dict:
                if genre in similarity_dict:
                    genre = similarity_dict[genre]
                write.writerow([user_id, time, genre, w])


def get_genre_list(fname):
    """
    edge_id, user_id, timestamp, tags

    0,user_000001,2006-08-13 14:59:59+00:00,"['electronic', 0.5319148936170213]"
    0,user_000001,2006-08-13 14:59:59+00:00,"['alternative', 0.46808510638297873]"
    1,user_000001,2006-08-13 15:36:22+00:00,"['electronic', 0.6410256410256411]"
    1,user_000001,2006-08-13 15:36:22+00:00,"['chillout', 0.358974358974359]"
    2,user_000001,2006-08-13 15:40:13+00:00,"['math rock', 1.0]"
    3,user_000001,2006-08-15 13:41:18+00:00,"['electronica', 1.0]"
    """
    edgelist = open(fname, "r")
    lines = list(edgelist.readlines())
    edgelist.close()
    genre_dict = {}
    for i in range(1, len(lines)):
        vals = lines[i].split(",")
        user_id = vals[1]
        time = vals[2]
        genre = vals[3].strip('"').strip("['")
        # genre = vals[3]
        w = float(vals[4][:-3])
        if genre not in genre_dict:
            genre_dict[genre] = 1
        else:
            genre_dict[genre] += 1

    # TODO check the frequency of genres and threshold
    genre_list_10 = []
    genre_list_100 = []
    genre_list_1000 = []
    genre_list_2000 = []
    for key, freq in genre_dict.items():
        if freq > 10:
            genre_list_10.append([key])
        if freq > 100:
            genre_list_100.append([key])
        if freq > 1000:
            genre_list_1000.append([key])
        if freq > 2000:
            genre_list_2000.append([key])
    print("number of genres with frequency > 10: " + str(len(genre_list_10)))
    print("number of genres with frequency > 100: " + str(len(genre_list_100)))
    print("number of genres with frequency > 1000: " + str(len(genre_list_1000)))
    print("number of genres with frequency > 2000: " + str(len(genre_list_2000)))
    fields = ["genre"]

    with open("genre_list_1000.csv", "w") as f:
        write = csv.writer(f)
        write.writerow(fields)
        write.writerows(genre_list_1000)


def find_unique_genres(fname: str, threshold: float = 0.8):
    """
    identify fuzzy strings which are actually the same genre, differences can be spacing, typo etc.
    """
    # load all genre names into a list
    edgelist = open(fname, "r")
    lines = list(edgelist.readlines())
    edgelist.close()

    genres = []
    sim_genres = {}
    for i in range(1, len(lines)):
        line = lines[i]
        genre = line.strip("\n")
        genres.append(genre)

    for i in range(len(genres)):
        for j in range(i + 1, len(genres)):
            text = genres[i]
            search_key = genres[j]
            sim = SequenceMatcher(None, text, search_key)
            sim = sim.ratio()
            if sim >= threshold:
                sim_genres[(text, search_key)] = sim

    print("there are " + str(len(sim_genres)) + " similar genres")
    print(sim_genres)


def load_genre_dict(
    fname: str,
) -> Dict[str, Any]:
    """
    reading the list of genres from genre_list.csv
    parameters:
        fname: file name of the genre list
    Returns:
        genre_dict: a dictionary of genres
    """
    genre_dict = {}
    with open(fname, "r") as f:
        reader = csv.reader(f)
        for row in reader:
            genre_dict[row[0]] = 1
    return genre_dict


def generate_daily_node_labels(fname: str):
    r"""
    read a temporal edgelist
    node label = fav genre in this day
    generate the node label for each day for each user
    Note: only genres from the genre_list are considered

    user_000001,2006-08-13 14:59:59+00:00,"['electronic', 0.5319148936170213]"
    user_000001,2006-08-13 14:59:59+00:00,"['alternative', 0.46808510638297873]"
    user_000001,2006-08-13 15:36:22+00:00,"['electronic', 0.6410256410256411]"
    user_000001,2006-08-13 15:36:22+00:00,"['chillout', 0.358974358974359]"
    user_000001,2006-08-13 15:40:13+00:00,"['math rock', 1.0]"
    user_000001,2006-08-15 13:41:18+00:00,"['electronica', 1.0]"
    user_000001,2006-08-15 13:59:27+00:00,"['acid jazz', 0.3546099290780142]"
    user_000001,2006-08-15 13:59:27+00:00,"['nu jazz', 0.3333333333333333]"
    user_000001,2006-08-15 13:59:27+00:00,"['chillout', 0.3120567375886525]"
    """
    edgelist = open(fname, "r")
    lines = list(edgelist.readlines())
    edgelist.close()

    format = "%Y-%m-%d %H:%M:%S"
    day_dict = {}  # store the weights of genres on this day
    cur_day = -1

    with open("daily_labels.csv", "w") as outf:
        write = csv.writer(outf)
        fields = ["user_id", "year", "month", "day", "genre", "weight"]
        write.writerow(fields)

        # generate daily labels for users
        for i in range(1, len(lines)):
            vals = lines[i].split(",")
            user_id = vals[0]
            time = vals[1][:-7]
            date_object = datetime.datetime.strptime(time, format)
            if i == 1:
                cur_day = date_object.day

            genre = vals[2]
            w = float(vals[3].strip())
            if date_object.day != cur_day:
                #! normalize the weights in the day_dict to sum 1
                # * remove normalization for future aggregation
                # total = sum(day_dict.values())
                # day_dict = {k: v / total for k, v in day_dict.items()}

                #! user,time,genre,weight  # genres = # of weights
                out = [
                    user_id,
                    str(date_object.year),
                    str(date_object.month),
                    str(date_object.day),
                ]
                for genre, w in day_dict.items():
                    write.writerow(out + [genre] + [w])

                cur_day = date_object.day
                day_dict = {}
            else:
                if genre not in day_dict:
                    day_dict[genre] = w
                else:
                    day_dict[genre] += w


def generate_aggregate_labels(fname: str, days: int = 7):
    """
    aggregate the genres over a number of days,  as specified by days
    #! current generation includes edges from the day of the label, thus the label should be set to be beginning of the day
    prediction should always be at the first second of the day
    """
    edgelist = open(fname, "r")
    lines = list(edgelist.readlines())
    edgelist.close()
    date_prev = 0

    genre_dict = {}
    user_prev = 0

    # "user_id", "year", "month", "day", "genre", "weight"
    with open(str(days) + "days_labels.csv", "w") as outf:
        write = csv.writer(outf)
        fields = ["user_id", "year", "month", "day", "genre", "weight"]
        write.writerow(fields)

        for i in range(1, len(lines)):
            vals = lines[i].split(",")
            user_id = vals[0]
            year = int(vals[1])
            month = int(vals[2])
            day = int(vals[3])
            genre = vals[4]
            w = float(vals[5])
            if i == 1:
                date_prev = date(year, month, day)
                user_prev = user_id

            date_cur = date(year, month, day)

            if user_id != user_prev:
                date_prev = date(year, month, day)
                user_prev = user_id

            if (
                date_cur - date_prev
            ).days <= days:  #! this means that the date = [0,7] which includes the current day
                if genre not in genre_dict:
                    genre_dict[genre] = w
                else:
                    genre_dict[genre] += w
            else:
                # start a new week
                # normalize the weight to sum 1
                total = sum(genre_dict.values())
                genre_dict = {k: v / total for k, v in genre_dict.items()}

                out = [
                    user_id,
                    str(date_prev.year),
                    str(date_prev.month),
                    str(date_prev.day),
                ]
                for genre, w in genre_dict.items():
                    write.writerow(out + [genre] + [w])
                date_prev = date_prev + datetime.timedelta(days=1)
                genre_dict = {}


def most_frequent(List):
    """
    helper function to find the most frequent element in a list
    the ties are broken by choosing the earlier element
    """
    counter = 0
    out = List[0]

    for item in List:
        curr_frequency = List.count(item)
        if curr_frequency > counter:  # update on most frequent item is found
            counter = curr_frequency
            out = item
    return out


def convert_ts_unix(fname: str, outname: str):
    """
    convert all time from datetime to unix time
    """
    TIME_FORMAT = "%Y-%m-%d"
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "user_id", "genre", "weight"]
        write.writerow(fields)

        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            # time,user_id,genre,weight
            for row in csv_reader:
                if line_count == 0:
                    line_count += 1
                else:
                    ts = datetime.datetime.strptime(row[0], TIME_FORMAT)
                    ts += timedelta(days=1)
                    ts = int(ts.timestamp())
                    user_id = row[1]
                    genre = row[2]
                    weight = float(row[3])
                    write.writerow([ts, user_id, genre, weight])


def convert_ts_edgelist(fname: str, outname: str):
    """
    convert all time from datetime to unix time
    """
    TIME_FORMAT = "%Y-%m-%d %H:%M:%S"
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "user_id", "genre", "weight"]
        write.writerow(fields)

        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            # time,user_id,genre,weight
            for row in csv_reader:
                if line_count == 0:
                    line_count += 1
                else:
                    ts = datetime.datetime.strptime(row[0], TIME_FORMAT)
                    ts = int(ts.timestamp())
                    user_id = row[1]
                    genre = row[2]
                    weight = float(row[3])
                    write.writerow([ts, user_id, genre, weight])


def sort_node_labels(fname, outname):
    r"""
    sort the node labels by time
    """
    edgelist = open(fname, "r")
    lines = list(edgelist.readlines())
    edgelist.close()

    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["time", "user_id", "genre", "weight"]
        write.writerow(fields)
        rows_dict = {}

        for i in range(1, len(lines)):
            vals = lines[i].split(",")
            user_id = vals[0]
            year = int(vals[1])
            month = int(vals[2])
            day = int(vals[3])
            genre = vals[4]
            w = float(vals[5])
            date_cur = datetime(year, month, day)
            time_ts = date_cur.strftime("%Y-%m-%d")
            if time_ts not in rows_dict:
                rows_dict[time_ts] = [(user_id, genre, w)]
            else:
                rows_dict[time_ts].append((user_id, genre, w))

        time_keys = list(rows_dict.keys())
        time_keys.sort()

        for ts in time_keys:
            rows = rows_dict[ts]
            for user_id, genre, w in rows:
                write.writerow([ts, user_id, genre, w])


def sort_edgelist(fname, outname="sorted_lastfm_edgelist.csv"):
    r"""
    sort the edgelist by time
    """
    edgelist = open(fname, "r")
    lines = list(edgelist.readlines())
    edgelist.close()

    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["time", "user_id", "genre", "weight"]
        write.writerow(fields)

        rows_dict = {}
        for idx in range(1, len(lines)):
            vals = lines[idx].split(",")
            user_id = vals[0]
            time_ts = vals[1][:-7]
            genre = vals[2]
            w = float(vals[3].strip())
            if time_ts not in rows_dict:
                rows_dict[time_ts] = [(user_id, genre, w)]
            else:
                rows_dict[time_ts].append((user_id, genre, w))

        time_keys = list(rows_dict.keys())
        time_keys.sort()

        for ts in time_keys:
            rows = rows_dict[ts]
            for user_id, genre, w in rows:
                write.writerow([ts, user_id, genre, w])


if __name__ == "__main__":
    #! generate the list of genres by frequency
    # get_genre_list("/mnt/c/Users/sheny/Desktop/TGB/tgb/datasets/lastfmGenre/dataset.csv")
    # genre_dict = load_genre_dict("/mnt/c/Users/sheny/Desktop/TGB/tgb/datasets/lastfmGenre/genre_list.csv")

    #! find similar genres
    # find_unique_genres("genre_list_1000.csv",threshold= 0.8)

    #! filter edgelist with genres to keep
    # genres_dict = load_genre_dict("genre_list_1000.csv")
    # filter_genre_edgelist("dataset.csv", genres_dict)

    #! generate the daily node labels
    # generate_daily_node_labels("lastfm_edgelist_clean.csv")

    # generate_daily_node_labels("/mnt/c/Users/sheny/Desktop/TGB/tgb/datasets/lastfmGenre/dataset.csv")
    # load_node_labels("/mnt/c/Users/sheny/Desktop/TGB/tgb/datasets/lastfmGenre/daily_labels.csv")

    # #! generate normalized weekly node labels
    # generate_aggregate_labels("daily_labels.csv", days=7)\

    # """
    # sort edgelist by time for lastfm dataset
    # """
    # fname = "../datasets/lastfmGenre/lastfm_edgelist_clean.csv"
    # outname = '../datasets/lastfmGenre/sorted_lastfm_edgelist.csv'
    # sort_edgelist(fname,
    #               outname = outname)

    # """
    # sort node labels by time for lastfm dataset
    # """
    # fname = "../datasets/lastfmGenre/7days_labels.csv"
    # outname = '../datasets/lastfmGenre/sorted_7days_node_labels.csv'
    # sort_node_labels(fname,
    #                  outname)

    # #! convert from date to ts
    # convert_ts_unix("lastfmgenre_node_labels_datetime.csv",
    #                 "lastfmgenre_node_labels.csv")
    convert_ts_edgelist("lastfmgenre_edgelist.csv", "lastfmgenre_edgelist_ts.csv")



# -*- coding: utf-8 -*-
# @Time    : 2019/12/5 4:20 下午
# @Author  : Lee_zix
# @Email   : Lee_zix@163.com
# @File    : ent2word.py.py
# @Software: PyCharm

import os

def load_index(input_path):
    index, rev_index = {}, {}
    with open(input_path) as f:
        for i, line in enumerate(f.readlines()):        # relaions.dict和entities.dict中的id都是按顺序排列的
            rel, id = line.strip().split("\t")
            index[rel] = id
            rev_index[id] = rel
    return index, rev_index

entity2id, id2entity = load_index(os.path.join('entity2id.txt'))
relation2id, id2relation = load_index(os.path.join('relation2id.txt'))

count = 0
count1 = 0
word_list = set()
for entity_str in entity2id.keys():
    if "(" in entity_str and ")" in entity_str:
        count += 1
        begin = entity_str.find('(')
        end = entity_str.find(')')
        w1 = entity_str[:begin].strip()
        w2 = entity_str[begin+1: end]
        if w2 not in entity2id.keys():
            print(w2)
            count1 += 1
        word_list.add(w1)
        word_list.add(w2)
    else:
        word_list.add(entity_str)

num_word = len(word_list)

word2id = {word: id for id, word in enumerate(word_list)}
id2word = {id: word for id, word in enumerate(word_list)}
# print(word2id)
# print(id2word)

print("words num: {}, enity_num: {}".format(num_word, len(entity2id.keys())))
print(float(count)/len(entity2id.keys()))
print(float(count1)/float(count))

with open("word2id.txt", "w") as f:
    for word in word2id.keys():
        f.write(word + "\t" + str(word2id[word])+'\n')

eid2wid = []
for id in range(len(id2entity.keys())):
    entity_str = id2entity[str(id)]
    if "(" in entity_str and ")" in entity_str:
        count += 1
        begin = entity_str.find('(')
        end = entity_str.find(')')
        w1 = entity_str[:begin].strip()
        w2 = entity_str[begin+1: end]
        eid2wid.append([str(entity2id[entity_str]), "0", str(word2id[w1])])   # isA关系
        eid2wid.append([str(entity2id[entity_str]), "1", str(word2id[w2])])     # 隶属关系
    else:
        eid2wid.append([str(entity2id[entity_str]), "2", str(word2id[entity_str])])

with open("e-w-graph.txt", "w") as f:
    for line in eid2wid:
        f.write("\t".join(line)+'\n')







"""
Evaluator Module for Dynamic Link Prediction
"""

import numpy as np
from sklearn.metrics import *
import math
from tgb.utils.info import DATA_EVAL_METRIC_DICT



try:
    import torch
except ImportError:
    torch = None


class Evaluator(object):
    r"""Evaluator for Link Property Prediction """

    def __init__(self, name: str, k_value: int = 10):
        r"""
        Parameters:
            name: name of the dataset
            k_value: the desired 'k' value for calculating metric@k
        """
        self.name = name
        self.k_value = k_value  # for computing `hits@k`
        self.valid_metric_list = ['hits@', 'mrr']
        if self.name not in DATA_EVAL_METRIC_DICT:
            raise NotImplementedError("Dataset not supported")
    
    def _parse_and_check_input(self, input_dict):
        r"""
        Check whether the input has the appropriate format
        Parametrers:
            input_dict: a dictionary containing "y_pred_pos", "y_pred_neg", and "eval_metric"
            note: "eval_metric" should be a list including one or more of the followin metrics: ["hits@", "mrr"]
        Returns:
            y_pred_pos: positive predicted scores
            y_pred_neg: negative predicted scores
        """

        if "eval_metric" not in input_dict:
            raise RuntimeError("Missing key of eval_metric!")

        for eval_metric in input_dict["eval_metric"]:
            if eval_metric in self.valid_metric_list:
                if "y_pred_pos" not in input_dict:
                    raise RuntimeError("Missing key of y_true")
                if "y_pred_neg" not in input_dict:
                    raise RuntimeError("Missing key of y_pred")

                y_pred_pos, y_pred_neg = input_dict["y_pred_pos"], input_dict["y_pred_neg"]

                # converting to numpy on cpu
                if torch is not None and isinstance(y_pred_pos, torch.Tensor):
                    y_pred_pos = y_pred_pos.detach().cpu().numpy()
                if torch is not None and isinstance(y_pred_neg, torch.Tensor):
                    y_pred_neg = y_pred_neg.detach().cpu().numpy()

                # check type and shape
                if not isinstance(y_pred_pos, np.ndarray) or not isinstance(y_pred_neg, np.ndarray):
                    raise RuntimeError(
                        "Arguments to Evaluator need to be either numpy ndarray or torch tensor!"
                    )
            else:
                print(
                    "ERROR: The evaluation metric should be in:", self.valid_metric_list
                )
                raise ValueError("Unsupported eval metric %s " % (eval_metric))
        self.eval_metric = input_dict["eval_metric"]

        return y_pred_pos, y_pred_neg

    def _eval_hits_and_mrr(self, y_pred_pos, y_pred_neg, type_info, k_value):
        r"""
        compute hist@k and mrr
        reference:
            - https://github.com/snap-stanford/ogb/blob/d5c11d91c9e1c22ed090a2e0bbda3fe357de66e7/ogb/linkproppred/evaluate.py#L214
        
        Parameters:
            y_pred_pos: positive predicted scores
            y_pred_neg: negative predicted scores
            type_info: type of the predicted scores; could be 'torch' or 'numpy'
            k_value: the desired 'k' value for calculating metric@k
        
        Returns:
            a dictionary containing the computed performance metrics
        """
        if type_info == 'torch':
            # calculate ranks
            y_pred_pos = y_pred_pos.view(-1, 1)
            # optimistic rank: "how many negatives have a larger score than the positive?"
            # ~> the positive is ranked first among those with equal score
            optimistic_rank = (y_pred_neg > y_pred_pos).sum(dim=1)
            # pessimistic rank: "how many negatives have at least the positive score?"
            # ~> the positive is ranked last among those with equal score
            pessimistic_rank = (y_pred_neg >= y_pred_pos).sum(dim=1)
            ranking_list = 0.5 * (optimistic_rank + pessimistic_rank) + 1
            hitsK_list = (ranking_list <= k_value).to(torch.float)
            mrr_list = 1./ranking_list.to(torch.float)

            return {
                    f'hits@{k_value}': hitsK_list.mean(),
                    'mrr': mrr_list.mean()
                    }

        else:
            y_pred_pos = y_pred_pos.reshape(-1, 1)
            optimistic_rank = (y_pred_neg > y_pred_pos).sum(axis=1)
            pessimistic_rank = (y_pred_neg >= y_pred_pos).sum(axis=1)
            ranking_list = 0.5 * (optimistic_rank + pessimistic_rank) + 1
            hitsK_list = (ranking_list <= k_value).astype(np.float32)
            mrr_list = 1./ranking_list.astype(np.float32)

            return {
                    f'hits@{k_value}': hitsK_list.mean(),
                    'mrr': mrr_list.mean()
                    }

    def eval(self, 
             input_dict: dict, 
             verbose: bool = False) -> dict:
        r"""
        evaluate the link prediction task
        this method is callable through an instance of this object to compute the metric

        Parameters:
            input_dict: a dictionary containing "y_pred_pos", "y_pred_neg", and "eval_metric"
                        the performance metric is calculated for the provided scores
            verbose: whether to print out the computed metric
        
        Returns:
            perf_dict: a dictionary containing the computed performance metric
        """
        y_pred_pos, y_pred_neg = self._parse_and_check_input(input_dict)  # convert the predictions to numpy
        perf_dict = self._eval_hits_and_mrr(y_pred_pos, y_pred_neg, type_info='numpy', k_value=self.k_value)
        
        return perf_dict
    



import csv

def load_index(input_path):
    index, rev_index = {}, {}
    with open(input_path) as f:
        for i, line in enumerate(f.readlines()):        # relaions.dict和entities.dict中的id都是按顺序排列的
            rel, id = line.strip().split("\t")
            index[rel] = id
            rev_index[id] = rel
    return index, rev_index


def load_tab_list(input_path):
    rows = []
    with open(input_path) as f:
        for i, line in enumerate(f.readlines()): 
            head,relation,tail,t, = line.strip().split("\t")
            rows.append([t,head,tail,relation])
    return rows

        
def write2csv(rows, output_path):
    with open(output_path, "w") as f:
        writer = csv.writer(f)
        writer.writerow(["timestamp", "head", "tail", "relation_type"])
        writer.writerows(rows)


def main():
    """
    concatenate and merge the edgelists into one 
    change tab to ,
    """
    train_name = "train.txt"
    train_rows = load_tab_list(train_name)

    val_name = "valid.txt"
    val_rows = load_tab_list(val_name)

    test_name = "test.txt"
    test_rows = load_tab_list(test_name)

    all_rows = train_rows + val_rows + test_rows
    output_path = "icews14.csv"
    write2csv(all_rows, output_path)







if __name__ == "__main__":
    main()




import random
import os
import pickle
import sys
import argparse
import json
import torch
from typing import Any
import numpy as np
from torch_geometric.data import TemporalData
import pandas as pd
import torch


def add_inverse_quadruples(df: pd.DataFrame) -> pd.DataFrame:
    r"""
    adds the inverse relations required for the model to the dataframe
    """
    if ("edge_type" not in df):
        raise ValueError("edge_type is required to invert relation in TKG")
    
    sources = np.array(df["u"])
    destinations = np.array(df["i"])
    timestamps = np.array(df["ts"])
    edge_idxs = np.array(df["idx"])
    weights = np.array(df["w"])
    edge_type = np.array(df["edge_type"])

    num_rels = np.unique(edge_type).shape[0]
    inv_edge_type = edge_type + num_rels

    all_sources = np.concatenate([sources, destinations])
    all_destinations = np.concatenate([destinations, sources])
    all_timestamps = np.concatenate([timestamps, timestamps])
    all_edge_idxs = np.concatenate([edge_idxs, edge_idxs+edge_idxs.max()+1])
    all_weights = np.concatenate([weights, weights])
    all_edge_types = np.concatenate([edge_type, inv_edge_type])

    return pd.DataFrame(
            {
                "u": all_sources,
                "i": all_destinations,
                "ts": all_timestamps,
                "label": np.ones(all_timestamps.shape[0]),
                "idx": all_edge_idxs,
                "w": all_weights,
                "edge_type": all_edge_types,
            }
        )



def add_inverse_quadruples_np(quadruples: np.array, 
                              num_rels:int) -> np.array:
    """
    creates an inverse quadruple for each quadruple in quadruples. inverse quadruple swaps subject and objsect, and increases 
    relation id by num_rels
    :param quadruples: [np.array] dataset quadruples, [src, relation_id, dst, timestamp ]
    :param num_rels: [int] number of relations that we have originally
    returns all_quadruples: [np.array] quadruples including inverse quadruples
    """
    inverse_quadruples = quadruples[:, [2, 1, 0, 3]]
    inverse_quadruples[:, 1] = inverse_quadruples[:, 1] + num_rels  # we also need inverse quadruples
    all_quadruples = np.concatenate((quadruples[:,0:4], inverse_quadruples))
    return all_quadruples


def add_inverse_quadruples_pyg(data: TemporalData, num_rels:int=-1) -> list:
    r"""
    creates an inverse quadruple from PyG TemporalData object, returns both the original and inverse quadruples
    """
    timestamp = data.t
    head = data.src
    tail = data.dst
    msg = data.msg
    edge_type = data.edge_type #relation
    num_rels = torch.max(edge_type).item() + 1
    inv_type = edge_type + num_rels
    all_data = TemporalData(src=torch.cat([head, tail]), 
                            dst=torch.cat([tail, head]), 
                            t=torch.cat([timestamp, timestamp.clone()]), 
                            edge_type=torch.cat([edge_type, inv_type]), 
                            msg=torch.cat([msg, msg.clone()]),
                            y = torch.cat([data.y, data.y.clone()]),)
    return all_data



# import torch
def save_pkl(obj: Any, fname: str) -> None:
    r"""
    save a python object as a pickle file
    """
    with open(fname, "wb") as handle:
        pickle.dump(obj, handle, protocol=pickle.HIGHEST_PROTOCOL)


def load_pkl(fname: str) -> Any:
    r"""
    load a python object from a pickle file
    """
    with open(fname, "rb") as handle:
        return pickle.load(handle)

def set_random_seed(random_seed: int):
    r"""
    set random seed for reproducibility
    Args:
        random_seed (int): random seed
    """
    random.seed(random_seed)
    np.random.seed(random_seed)
    torch.manual_seed(random_seed)
    torch.cuda.manual_seed(random_seed)
    torch.cuda.manual_seed_all(random_seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    print(f'INFO: fixed random seed: {random_seed}')



def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]


def get_args():
    parser = argparse.ArgumentParser('*** TGB ***')
    parser.add_argument('-d', '--data', type=str, help='Dataset name', default='tgbl-wiki')
    parser.add_argument('--lr', type=float, help='Learning rate', default=1e-4)
    parser.add_argument('--bs', type=int, help='Batch size', default=200)
    parser.add_argument('--k_value', type=int, help='k_value for computing ranking metrics', default=10)
    parser.add_argument('--num_epoch', type=int, help='Number of epochs', default=30)
    parser.add_argument('--seed', type=int, help='Random seed', default=1)
    parser.add_argument('--mem_dim', type=int, help='Memory dimension', default=100)
    parser.add_argument('--time_dim', type=int, help='Time dimension', default=100)
    parser.add_argument('--emb_dim', type=int, help='Embedding dimension', default=100)
    parser.add_argument('--tolerance', type=float, help='Early stopper tolerance', default=1e-6)
    parser.add_argument('--patience', type=float, help='Early stopper patience', default=5)
    parser.add_argument('--num_run', type=int, help='Number of iteration runs', default=5)

    try:
        args = parser.parse_args()
    except:
        parser.print_help()
        sys.exit(0)
    return args, sys.argv




def save_results(new_results: dict, filename: str):
    r"""
    save (new) results into a json file
    :param: new_results (dictionary): a dictionary of new results to be saved
    :filename: the name of the file to save the (new) results
    """
    if os.path.isfile(filename):
        # append to the file
        with open(filename, 'r+') as json_file:
            file_data = json.load(json_file)
            # convert file_data to list if not
            if type(file_data) is dict:
                file_data = [file_data]
            file_data.append(new_results)
            json_file.seek(0)
            json.dump(file_data, json_file, indent=4)
    else:
        # dump the results
        with open(filename, 'w') as json_file:
            json.dump(new_results, json_file, indent=4)


def split_by_time(data):
    """
    https://github.com/Lee-zix/CEN/blob/main/rgcn/utils.py
    create list where each entry has an entry with all triples for this timestep
    """
    timesteps = list(set(data[:,3]))
    timesteps.sort()
    snapshot_list = [None] * len(timesteps)

    for index, ts in enumerate(timesteps):
        mask = np.where(data[:, 3] == ts)[0]
        snapshot_list[index] = data[mask,:3]

    return snapshot_list

